Package: diffcp
Type: Package
Title: Differentiating Through Cone Programs
Version: 0.1.2
Authors@R: c(
    person("Balasubramanian", "Narasimhan",
           role = c("aut", "cre"),
           email = "naras@stanford.edu"),
    person("Akshay", "Agrawal", role = "aut"),
    person("Shane", "Barratt",  role = "aut"),
    person("Stephen", "Boyd",   role = "aut"),
    person("Enzo", "Busseti",   role = "aut"),
    person("Walaa", "Moursi",   role = "aut"))
Description: A port of the 'python' 'diffcp' package. Computes the
    derivative of the optimal solution map of a convex cone program,
    treating the program as an implicit function of its data
    (constraint matrix, offset, objective coefficients, and optionally
    a quadratic), mirroring Agrawal et al. (2019)
    <doi:10.48550/arXiv.1904.09043>.
URL: https://bnaras.github.io/diffcp/
BugReports: https://github.com/bnaras/diffcp/issues
License: Apache License (>= 2)
Encoding: UTF-8
Imports: cli, clarabel, Matrix, methods, Rcpp
LinkingTo: Rcpp, RcppEigen
Suggests: knitr, pkgdown, rmarkdown, scs, testthat (>= 3.0.0)
VignetteBuilder: knitr
SystemRequirements: C++17
Config/testthat/edition: 3
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-08-22 20:42:06 UTC; naras
Author: Balasubramanian Narasimhan [aut, cre],
  Akshay Agrawal [aut],
  Shane Barratt [aut],
  Stephen Boyd [aut],
  Enzo Busseti [aut],
  Walaa Moursi [aut]
Maintainer: Balasubramanian Narasimhan <naras@stanford.edu>
Repository: CRAN
Date/Publication: 2026-08-22 23:20:24 UTC

# Generated by roxygen2: do not edit by hand

export(CONES)
export(EQ_DIM)
export(EXP)
export(EXP_DUAL)
export(POS)
export(PSD)
export(SOC)
export(ZERO)
export(parse_cone_dict)
export(pi)
export(solve_and_derivative)
export(solve_only)
export(unvec_symm)
export(upstream_info)
export(vec_symm)
import(Matrix)
importFrom(Rcpp,evalCpp)
useDynLib(diffcp, .registration = TRUE)

[![Build Status](http://github.com/cvxgrp/diffcp/workflows/build/badge.svg?event=push)](https://github.com/cvxgrp/diffcp/actions/workflows/build.yml)

# diffcp

`diffcp` is a Python package for computing the derivative of a convex cone program, with respect to its problem data. The derivative is implemented as an abstract linear map, with methods for its forward application and its adjoint. 

The implementation is based on the calculations in our paper [Differentiating through a cone program](http://web.stanford.edu/~boyd/papers/diff_cone_prog.html).

### Installation
`diffcp` is available on PyPI, as a source distribution. Install it with

```bash
pip install diffcp
```

You will need a C++11-capable compiler to build `diffcp`.

`diffcp` requires:
* [NumPy](https://github.com/numpy/numpy) >= 1.15
* [SciPy](https://github.com/scipy/scipy) >= 1.10
* [SCS](https://github.com/bodono/scs-python) >= 2.0.2
* [pybind11](https://github.com/pybind/pybind11/tree/stable) >= 2.4
* [threadpoolctl](https://github.com/joblib/threadpoolctl) >= 1.1
* [ECOS](https://github.com/embotech/ecos-python) >= 2.0.10
* Python >= 3.7

`diffcp` uses Eigen; Eigen operations can be automatically vectorized by compilers. To enable vectorization, install with

```bash
MARCH_NATIVE=1 pip install diffcp
```

OpenMP can be enabled by passing extra arguments to your compiler. For example, on linux, you can tell gcc to activate the OpenMP extension by specifying the flag "-fopenmp":

```bash
OPENMP_FLAG="-fopenmp" pip install diffcp
```

To enable both vectorization and OpenMP (on linux), use

```bash
MARCH_NATIVE=1 OPENMP_FLAG="-fopenmp" pip install diffcp
```

### Cone programs
`diffcp` differentiates through a primal-dual cone program pair. The primal problem must be expressed as 

```
minimize        c'x
subject to      Ax + s = b
                s in K
```
where  `x` and `s` are variables, `A`, `b` and `c` are the user-supplied problem data, and `K` is a user-defined convex cone. The corresponding dual problem is

```
minimize        b'y
subject to      A'y + c == 0
                y in K^*
```

with dual variable `y`.

### Usage

`diffcp` exposes the function

```python
solve_and_derivative(A, b, c, cone_dict, warm_start=None, solver=None, **kwargs).
```

This function returns a primal-dual solution `x`, `y`, and `s`, along with
functions for evaluating the derivative and its adjoint (transpose).
These functions respectively compute right and left multiplication of the derivative
of the solution map at `A`, `b`, and `c` by a vector.
The `solver` argument determines which solver to use; the available solvers
are `solver="SCS"` and `solver="ECOS"`.
If no solver is specified, `diffcp` will choose the solver itself.
In the case that the problem is not solved, i.e. the solver fails for some reason, we will raise
a `SolverError` Exception.

#### Arguments
The arguments `A`, `b`, and `c` correspond to the problem data of a cone program.
* `A` must be a [SciPy sparse CSC matrix](https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csc_matrix.html).
* `b` and `c` must be NumPy arrays.
* `cone_dict` is a dictionary that defines the convex cone `K`.
* `warm_start` is an optional tuple `(x, y, s)` at which to warm-start. (Note: this is only available for the SCS solver).
* `**kwargs` are keyword arguments to forward to the solver (e.g., `verbose=False`).

These inputs must conform to the [SCS convention](https://github.com/bodono/scs-python) for problem data. The keys in `cone_dict` correspond to the cones, with
* `diffcp.ZERO` for the zero cone,
* `diffcp.POS` for the positive orthant,
* `diffcp.SOC` for a product of SOC cones,
* `diffcp.PSD` for a product of PSD cones, and
* `diffcp.EXP` for a product of exponential cones.

The values in `cone_dict` denote the sizes of each cone; the values of `diffcp.SOC`, `diffcp.PSD`, and `diffcp.EXP` should be lists. The order of the rows of `A` must match the ordering of the cones given above. For more details, consult the [SCS documentation](https://github.com/cvxgrp/scs/blob/master/README.md).

#### Return value
The function `solve_and_derivative` returns a tuple

```python
(x, y, s, derivative, adjoint_derivative)
```

* `x`, `y`, and `s` are a primal-dual solution.

* `derivative` is a function that applies the derivative at `(A, b, c)` to perturbations `dA`, `db`, `dc`. It has the signature 
```derivative(dA, db, dc) -> dx, dy, ds```, where `dA` is a SciPy sparse CSC matrix with the same sparsity pattern as `A`, and `db` and `dc` are NumPy arrays. `dx`, `dy`, and `ds` are NumPy arrays, approximating the change in the primal-dual solution due to the perturbation.

* `adjoint_derivative` is a function that applies the adjoint of the derivative to perturbations `dx`, `dy`, `ds`. It has the signature 
```adjoint_derivative(dx, dy, ds) -> dA, db, dc```, where `dx`, `dy`, and `ds` are NumPy arrays.

#### Example
```python
import numpy as np
from scipy import sparse

import diffcp

cone_dict = {
    diffcp.ZERO: 3,
    diffcp.POS: 3,
    diffcp.SOC: [5]
}

m = 3 + 3 + 5
n = 5

A, b, c = diffcp.utils.random_cone_prog(m, n, cone_dict)
x, y, s, D, DT = diffcp.solve_and_derivative(A, b, c, cone_dict)

# evaluate the derivative
nonzeros = A.nonzero()
data = 1e-4 * np.random.randn(A.size)
dA = sparse.csc_matrix((data, nonzeros), shape=A.shape)
db = 1e-4 * np.random.randn(m)
dc = 1e-4 * np.random.randn(n)
dx, dy, ds = D(dA, db, dc)

# evaluate the adjoint of the derivative
dx = c
dy = np.zeros(m)
ds = np.zeros(m)
dA, db, dc = DT(dx, dy, ds)
```

For more examples, including the SDP example described in the paper, see the [`examples`](examples/) directory.

### Citing
If you wish to cite `diffcp`, please use the following BibTex:

```
@article{diffcp2019,
    author       = {Agrawal, A. and Barratt, S. and Boyd, S. and Busseti, E. and Moursi, W.},
    title        = {Differentiating through a Cone Program},
    journal      = {Journal of Applied and Numerical Optimization},
    year         = {2019},
    volume       = {1},
    number       = {2},
    pages        = {107--115},
}

@misc{diffcp,
    author       = {Agrawal, A. and Barratt, S. and Boyd, S. and Busseti, E. and Moursi, W.},
    title        = {{diffcp}: differentiating through a cone program, version 1.0},
    howpublished = {\url{https://github.com/cvxgrp/diffcp}},
    year         = 2019
}
```

The following thesis concurrently derived the mathematics behind differentiating cone programs.
```
@phdthesis{amos2019differentiable,
  author       = {Brandon Amos},
  title        = {{Differentiable Optimization-Based Modeling for Machine Learning}},
  school       = {Carnegie Mellon University},
  year         = 2019,
  month        = May,
}
```

## DIFFCP SOURCE: diffcp/cpp/src/dprojection.cpp @ v1.1.8 (6a78143) (C++ extension).
## Upstream pin: see inst/UPSTREAM.dcf / upstream_info().
##
## Per-cone Jacobians of the projection operator.
##
## For each cone K, dpi_K(x) is the (sub-)derivative of Pi_K at x: a
## linear operator from R^n -> R^n. We expose each as a *closure*
## with `matvec(v)` and `rmatvec(v)` methods (the projection's
## Jacobian is symmetric for self-dual cones, so matvec == rmatvec
## there; we still expose both for uniformity with the more general
## non-self-dual cases.)

# -- Zero cone ----------------------------------------------------
## Pi_0(x) = 0 (primal). dPi = 0.
## Pi_{R^n}(x) = x (dual). dPi = I.
.dproj_zero <- function(x, dual = FALSE) {
  n <- length(x)
  if (dual) {
    ## identity
    list(
      matvec  = function(v) v,
      rmatvec = function(v) v,
      n = n
    )
  } else {
    ## zero
    list(
      matvec  = function(v) numeric(n),
      rmatvec = function(v) numeric(n),
      n = n
    )
  }
}

# -- Nonneg orthant -----------------------------------------------
## Pi_{R_+}(x)_i = max(x_i, 0). dPi = diag(1[x > 0]).
## Self-dual: same Jacobian for primal and dual.
.dproj_nonneg <- function(x, dual = FALSE) {
  d <- as.numeric(x > 0)
  list(
    matvec  = function(v) d * v,
    rmatvec = function(v) d * v,
    n = length(x)
  )
}

# -- Single SOC -----------------------------------------------------
## Q^{n+1} = { (t, z) : ||z|| <= t }. Self-dual.
##
## DIFFCP SOURCE: diffcp/cpp/src/dprojection.cpp (SOC branch).
## At x = (t, z), let n = ||z||.
##   - If x is in interior of K (n < t): dPi = I.
##   - If x is in interior of -K (n < -t): dPi = 0.
##   - On the boundary or in the "transition" region, dPi has the
##     closed form below (a low-rank update of a smoothed identity).
##
## The non-trivial branch (norm_z > |t|): the projection writes
##   Pi(x) = (alpha, alpha * z / n)  where alpha = (t + n) / 2.
## Differentiating gives a 2x2 block structure between t and the
## radial direction z/n, plus an isotropic component on the
## tangential complement.
.dproj_soc_one <- function(x, dual = FALSE) {
  n <- length(x)
  if (n == 0L) {
    return(list(matvec = function(v) v, rmatvec = function(v) v, n = 0L))
  }
  t <- x[1L]
  z <- x[-1L]
  norm_z <- sqrt(sum(z * z))

  if (norm_z <= t) {
    ## Interior of K -> identity
    return(list(
      matvec  = function(v) v,
      rmatvec = function(v) v,
      n = n
    ))
  }
  if (norm_z <= -t) {
    ## Interior of -K (= polar of K) -> zero
    return(list(
      matvec  = function(v) numeric(n),
      rmatvec = function(v) numeric(n),
      n = n
    ))
  }

  ## Boundary / transition region.
  ## D = 0.5 * (I + [t/||z|| 0; 0 (t + ||z||)/(2||z||) * I - t/(2||z||) * z z^T / ||z||^2])
  ## Hand-derived from differentiating the closed-form projection;
  ## standard reference is the SCS / diffcp source.
  s <- t / norm_z
  zhat <- z / norm_z
  ## Block matvec:
  ##   v0 := v[1], w := v[-1]
  ##   out0 = 0.5 * (v0 + s * (zhat . w))
  ##   out_w = 0.5 * (s * v0 * zhat + (1 + s) * w - s * (zhat . w) * zhat)
  matvec <- function(v) {
    v0 <- v[1L]
    w  <- v[-1L]
    zw <- sum(zhat * w)
    out0 <- 0.5 * (v0 + zw)
    out_w <- 0.5 * (v0 * zhat + (1 + s) * w - (1 - s) * zw * zhat)
    c(out0, out_w)
  }
  ## SOC projection's Jacobian is symmetric (self-dual cone), so
  ## rmatvec == matvec.
  list(matvec = matvec, rmatvec = matvec, n = n)
}

# -- Cartesian product Jacobian -----------------------------------
## Builds a block-diagonal Jacobian operator from per-cone-block
## Jacobians, returned as a closure with matvec / rmatvec.
##
## Mirrors `diffcp._diffcp.dprojection(v, cones_parsed, dual)` (Python
## C++ extension). Only Zero, Nonneg, and SOC are supported in this
## development version; PSD and EXP arrive in subsequent commits.
.dprojection <- function(x, cones, dual = FALSE) {
  blocks <- list()
  offset <- 0L
  for (entry in cones) {
    cone <- entry$name
    sz   <- entry$sizes
    if (length(sz) == 0L || sum(sz) == 0L) next
    if (cone == "q") {
      for (d in sz) {
        idx <- (offset + 1L):(offset + d)
        blocks[[length(blocks) + 1L]] <- list(
          idx  = idx,
          jac  = .dproj_soc_one(x[idx], dual = dual)
        )
        offset <- offset + d
      }
    } else if (cone == "s" || cone == "ep" || cone == "ed") {
      cli::cli_abort("Jacobian for cone {.val {cone}} not yet implemented.")
    } else {
      for (d in sz) {
        idx <- (offset + 1L):(offset + d)
        jac <- switch(cone,
          z = .dproj_zero(x[idx], dual = dual),
          l = .dproj_nonneg(x[idx], dual = dual),
          cli::cli_abort("Unknown cone {.val {cone}}.")
        )
        blocks[[length(blocks) + 1L]] <- list(idx = idx, jac = jac)
        offset <- offset + d
      }
    }
  }
  total <- offset

  matvec <- function(v) {
    out <- numeric(total)
    for (b in blocks) {
      out[b$idx] <- b$jac$matvec(v[b$idx])
    }
    out
  }
  rmatvec <- function(v) {
    out <- numeric(total)
    for (b in blocks) {
      out[b$idx] <- b$jac$rmatvec(v[b$idx])
    }
    out
  }
  list(matvec = matvec, rmatvec = rmatvec, n = total)
}

## DIFFCP SOURCE: diffcp/cones.py @ v1.1.8 (6a78143) (auxiliary utilities).
## Upstream pin: see inst/UPSTREAM.dcf / upstream_info().
##
## Small helpers shared across cone projection / Jacobian code.

#' Symmetric vectorization (SCS convention)
#'
#' Returns a vectorized representation of a symmetric matrix `X`,
#' with off-diagonal entries scaled by sqrt(2) to make the SCS-style
#' dot product `<vec_symm(A), vec_symm(B)> = trace(A B)` hold.
#'
#' Mirrors `diffcp.cones.vec_symm` (Python).
#'
#' @param X A symmetric matrix.
#' @returns A numeric vector of length `n*(n+1)/2`.
#' @export
vec_symm <- function(X) {
  X <- as.matrix(X)
  n <- nrow(X)
  if (ncol(X) != n) {
    cli::cli_abort("{.fn vec_symm} requires a square matrix; got {nrow(X)}x{ncol(X)}.")
  }
  Y <- X * sqrt(2)
  diag(Y) <- diag(X)
  ## Upper-triangular indices in row-major order (matches SCS).
  ## Python uses np.triu_indices(n) which returns (rows, cols) such that
  ## row <= col -- iterate column-major over upper triangle.
  idx <- which(upper.tri(Y, diag = TRUE), arr.ind = TRUE)
  ## Order so that we walk columns first within each row (i.e., for each
  ## row, all upper-triangle entries in that row in increasing col).
  ord <- order(idx[, 1L], idx[, 2L])
  Y[idx[ord, , drop = FALSE]]
}

#' Inverse of `vec_symm`
#'
#' @param x A numeric vector of length `n*(n+1)/2`.
#' @param dim The matrix dimension `n`.
#' @returns The corresponding `n x n` symmetric matrix.
#' @export
unvec_symm <- function(x, dim) {
  if (length(x) != dim * (dim + 1L) / 2L) {
    cli::cli_abort(
      "{.arg x} has length {length(x)}; expected {dim * (dim + 1L) / 2L}."
    )
  }
  X <- matrix(0, dim, dim)
  ## Walk upper triangle in row-major (matching vec_symm).
  k <- 1L
  for (i in seq_len(dim)) {
    for (j in i:dim) {
      X[i, j] <- x[k]
      k <- k + 1L
    }
  }
  X <- X + t(X)
  X[upper.tri(X)] <- X[upper.tri(X)] / sqrt(2)
  X[lower.tri(X)] <- X[lower.tri(X)] / sqrt(2)
  ## Diagonal already counted twice (once via X, once via t(X)); halve.
  diag(X) <- diag(X) / 2
  X
}

#' PSD cone vectorized dimension
#'
#' For a `dim x dim` symmetric matrix, returns the length of its
#' upper-triangular packed representation `dim*(dim+1)/2`.
#' @param dim Matrix dimension.
#' @returns Integer.
#' @noRd
.vec_psd_dim <- function(dim) {
  as.integer(dim * (dim + 1L) / 2L)
}

#' Inverse: from packed length back to matrix dimension.
#' @noRd
.psd_dim <- function(size) {
  ## solve k*(k+1)/2 = size -> k = (-1 + sqrt(1 + 8*size)) / 2
  k <- (-1 + sqrt(1 + 8 * size)) / 2
  as.integer(round(k))
}

## Faithful R ports of diffcp/tests/test_cone_prog_diff.py.
##
## Tests grouped by topic (matching the order in the Python file):
##   * cone-dimension utilities (vec_psd_dim, psd_dim)
##   * EXP-cone membership (in_exp, in_exp_dual)
##   * vec_symm / unvec_symm round-trips
##   * per-cone projection sanity (_proj_zero/pos/soc/psd/exp)
##   * per-cone Jacobian sanity (dproj_*) via FD
##   * pi() and dpi() composition over multi-cone product
##   * lsqr vs lstsq sanity
##   * infeasible-problem error
##   * dprojection_exp analytical sensitivity
##
## Same atol bounds as Python.  Fixtures requiring cvxpy reference
## values come from inst/python_parity/cone_diff.py via
## helper-cone-diff-data.R (auto-sourced by testthat).

skip_if_not_installed("clarabel")
skip_if_not_installed("scs")

# ---- helper -----------------------------------------------------
.csc <- function(A_i, A_p, A_x, dims) {
  Matrix::sparseMatrix(i = A_i, p = A_p - 1L, x = A_x, dims = dims, index1 = TRUE)
}

# === test_vec_psd_dim ============================================
test_that("vec_psd_dim(10) == 55", {
  ## Mirrors Python `cone_lib.vec_psd_dim(10) == 10*11/2`.
  expect_equal(diffcp:::.vec_psd_dim(10L), 55L)
})

# === test_psd_dim ================================================
test_that("psd_dim(vec_psd_dim(4096)) == 4096", {
  expect_equal(diffcp:::.psd_dim(diffcp:::.vec_psd_dim(4096L)), 4096L)
})

# === test_in_exp =================================================
test_that("in_exp matches Python", {
  expect_true(diffcp:::cpp_in_exp(c(0, 0, 1)))
  expect_true(diffcp:::cpp_in_exp(c(-1, 0, 0)))
  expect_true(diffcp:::cpp_in_exp(c(1, 1, 5)))
  expect_false(diffcp:::cpp_in_exp(c(1, 0, 0)))
  expect_false(diffcp:::cpp_in_exp(c(-1, -1, 1)))
  expect_false(diffcp:::cpp_in_exp(c(-1, 0, -1)))
})

# === test_in_exp_dual ============================================
test_that("in_exp_dual matches Python", {
  expect_true(diffcp:::cpp_in_exp_dual(c(0, 1, 1)))
  expect_true(diffcp:::cpp_in_exp_dual(c(-1, 1, 5)))
  expect_false(diffcp:::cpp_in_exp_dual(c(0, -1, 1)))
  expect_false(diffcp:::cpp_in_exp_dual(c(0, 1, -1)))
})

# === test_unvec_symm =============================================
test_that("unvec_symm(vec_symm(X)) == X for symmetric X", {
  set.seed(0)
  n <- 5L
  x <- matrix(rnorm(n * n), n, n)
  x <- x + t(x)
  expect_equal(diffcp::unvec_symm(diffcp::vec_symm(x), n), x,
               tolerance = 1e-12)
})

# === test_vec_symm ===============================================
test_that("vec_symm(unvec_symm(x)) == x", {
  set.seed(0)
  n <- 5L
  m <- (n * (n + 1L)) %/% 2L
  x <- rnorm(m)
  expect_equal(diffcp::vec_symm(diffcp::unvec_symm(x, n)), x,
               tolerance = 1e-12)
})

# === test_proj_zero ==============================================
test_that("_proj_zero: dual=TRUE returns x; dual=FALSE returns 0", {
  set.seed(0)
  n <- 100L
  for (k in 1:10) {
    x <- rnorm(n)
    cones <- list(list(name = "z", sizes = n))
    expect_equal(diffcp::pi(x, cones, dual = TRUE),  x,         tolerance = 1e-15)
    expect_equal(diffcp::pi(x, cones, dual = FALSE), numeric(n),tolerance = 1e-15)
  }
})

# === test_proj_pos ===============================================
test_that("_proj_pos: returns max(x, 0); self-dual", {
  set.seed(0)
  n <- 100L
  cones <- list(list(name = "l", sizes = n))
  for (k in 1:15) {
    x <- rnorm(n)
    p <- diffcp::pi(x, cones, dual = FALSE)
    expect_equal(p, pmax(x, 0), tolerance = 1e-15)
    expect_equal(p, diffcp::pi(x, cones, dual = TRUE), tolerance = 1e-15)
  }
})

# === test_proj_soc ===============================================
test_that("_proj_soc matches Python diffcp reference (15 random vectors)", {
  ## Pinned reference from inst/python_parity/cone_diff.py:emit_proj_soc.
  n <- 100L
  cones <- list(list(name = "q", sizes = n))
  for (k in 1:15) {
    x <- proj_soc_x[((k - 1L) * n + 1L):(k * n)]
    p_ref <- proj_soc_p[((k - 1L) * n + 1L):(k * n)]
    p <- diffcp::pi(x, cones, dual = FALSE)
    expect_lt(max(abs(p - p_ref)), 1e-12)
    ## SOC is self-dual.
    expect_lt(max(abs(p - diffcp::pi(x, cones, dual = TRUE))), 1e-15)
  }
})

# === test_proj_psd ===============================================
test_that("_proj_psd matches Python diffcp reference (15 random matrices)", {
  n <- 10L
  m <- (n * (n + 1L)) %/% 2L
  cones <- list(list(name = "s", sizes = n))
  for (k in 1:15) {
    x <- proj_psd_x[((k - 1L) * m + 1L):(k * m)]
    p_ref <- proj_psd_p[((k - 1L) * m + 1L):(k * m)]
    p <- diffcp::pi(x, cones, dual = FALSE)
    expect_lt(max(abs(p - p_ref)), 1e-10)
    ## PSD is self-dual.
    expect_lt(max(abs(p - diffcp::pi(x, cones, dual = TRUE))), 1e-12)
  }
})

# === test_proj_exp ===============================================
test_that("_proj_exp matches Python diffcp reference (15 random vectors)", {
  ## 9-dim = 3 EXP cones.
  cones_exp <- list(list(name = "ep", sizes = 3L))
  cones_exp_dual <- list(list(name = "ed", sizes = 3L))
  for (k in 1:15) {
    x <- proj_exp_x[((k - 1L) * 9L + 1L):(k * 9L)]
    p_ref <- proj_exp_p[((k - 1L) * 9L + 1L):(k * 9L)]
    p_dual_ref <- proj_exp_p_dual[((k - 1L) * 9L + 1L):(k * 9L)]
    expect_lt(max(abs(diffcp::pi(x, cones_exp, dual = FALSE) - p_ref)), 1e-10)
    expect_lt(max(abs(diffcp::pi(x, cones_exp_dual, dual = FALSE) - p_dual_ref)),
              1e-10)
  }
})

# === Helper for dproj_* finite-difference test (mirrors _test_dproj) -
.test_dproj <- function(cone, dual, n, tol = 1e-8) {
  x <- rnorm(n)
  dx <- 1e-6 * rnorm(n)
  cones <- list(cone)
  proj_x <- diffcp::pi(x,      cones, dual = dual)
  proj_z <- diffcp::pi(x + dx, cones, dual = dual)
  Dpi_dense <- diffcp:::cpp_dprojection_dense(x, cones, dual)
  expect_lt(max(abs(Dpi_dense %*% dx - (proj_z - proj_x))), tol)
}

# === test_dproj_zero =============================================
test_that("dproj_zero matches FD (10 random)", {
  set.seed(0)
  cone <- list(name = "z", sizes = 55L)
  for (k in 1:10) .test_dproj(cone, dual = TRUE,  n = 55L)
  for (k in 1:10) .test_dproj(cone, dual = FALSE, n = 55L)
})

# === test_dproj_pos ==============================================
test_that("dproj_pos matches FD (10 random)", {
  set.seed(0)
  cone <- list(name = "l", sizes = 55L)
  for (k in 1:10) .test_dproj(cone, dual = TRUE,  n = 55L)
  for (k in 1:10) .test_dproj(cone, dual = FALSE, n = 55L)
})

# === test_dproj_soc ==============================================
test_that("dproj_soc matches FD (10 random)", {
  set.seed(0)
  cone <- list(name = "q", sizes = 55L)
  for (k in 1:10) .test_dproj(cone, dual = TRUE,  n = 55L)
  for (k in 1:10) .test_dproj(cone, dual = FALSE, n = 55L)
})

# === test_dproj_psd ==============================================
test_that("dproj_psd matches FD (10 random)", {
  set.seed(0)
  ## n=55 packed -> matrix dim 10
  cone <- list(name = "s", sizes = diffcp:::.psd_dim(55L))
  for (k in 1:10) .test_dproj(cone, dual = TRUE,  n = 55L)
  for (k in 1:10) .test_dproj(cone, dual = FALSE, n = 55L)
})

# === test_dproj_exp ==============================================
test_that("dproj_exp matches FD (10 random, tol=1e-5)", {
  set.seed(0)
  cone <- list(name = "ep", sizes = 18L)
  for (k in 1:10) .test_dproj(cone, dual = TRUE,  n = 54L, tol = 1e-5)
  for (k in 1:10) .test_dproj(cone, dual = FALSE, n = 54L, tol = 1e-5)
})

# === test_dproj_exp_dual =========================================
test_that("dproj_exp_dual matches FD (10 random, tol=1e-5)", {
  set.seed(0)
  cone <- list(name = "ed", sizes = 18L)
  for (k in 1:10) .test_dproj(cone, dual = TRUE,  n = 54L, tol = 1e-5)
  for (k in 1:10) .test_dproj(cone, dual = FALSE, n = 54L, tol = 1e-5)
})

# === test_pi (multi-cone composition) ============================
test_that("pi composes per-block _proj across multi-cone product", {
  set.seed(0)
  for (rep in 1:10) {
    zero_dim <- sample.int(9L, 1L)
    pos_dim  <- sample.int(9L, 1L)
    soc_dim  <- sample.int(9L, sample.int(9L, 1L), replace = TRUE)
    psd_dim  <- sample.int(9L, sample.int(9L, 1L), replace = TRUE)
    exp_dim  <- sample.int(15L, 1L) + 2L  # 3..17

    cones <- list(
      list(name = "z",  sizes = zero_dim),
      list(name = "l",  sizes = pos_dim),
      list(name = "q",  sizes = as.integer(soc_dim)),
      list(name = "s",  sizes = as.integer(psd_dim)),
      list(name = "ep", sizes = exp_dim),
      list(name = "ed", sizes = exp_dim)
    )
    size <- zero_dim + pos_dim + sum(soc_dim) +
      sum(sapply(psd_dim, function(d) (d * (d + 1L)) %/% 2L)) +
      2L * 3L * exp_dim
    x <- rnorm(size)

    for (dual in c(FALSE, TRUE)) {
      proj <- diffcp::pi(x, cones, dual = dual)

      offset <- 0L
      ## zero block
      block_z <- diffcp::pi(x[(offset + 1L):(offset + zero_dim)],
                            list(list(name = "z", sizes = zero_dim)),
                            dual = dual)
      expect_lt(max(abs(proj[(offset + 1L):(offset + zero_dim)] - block_z)), 1e-12)
      offset <- offset + zero_dim

      ## pos block
      block_l <- diffcp::pi(x[(offset + 1L):(offset + pos_dim)],
                            list(list(name = "l", sizes = pos_dim)),
                            dual = dual)
      expect_lt(max(abs(proj[(offset + 1L):(offset + pos_dim)] - block_l)), 1e-12)
      offset <- offset + pos_dim

      ## soc blocks
      for (d in soc_dim) {
        block_q <- diffcp::pi(x[(offset + 1L):(offset + d)],
                              list(list(name = "q", sizes = as.integer(d))),
                              dual = dual)
        expect_lt(max(abs(proj[(offset + 1L):(offset + d)] - block_q)), 1e-12)
        offset <- offset + d
      }

      ## psd blocks
      for (d in psd_dim) {
        m <- (d * (d + 1L)) %/% 2L
        block_s <- diffcp::pi(x[(offset + 1L):(offset + m)],
                              list(list(name = "s", sizes = as.integer(d))),
                              dual = dual)
        expect_lt(max(abs(proj[(offset + 1L):(offset + m)] - block_s)), 1e-10)
        offset <- offset + m
      }

      ## exp block
      block_ep <- diffcp::pi(x[(offset + 1L):(offset + 3L * exp_dim)],
                             list(list(name = "ep", sizes = exp_dim)),
                             dual = dual)
      expect_lt(max(abs(proj[(offset + 1L):(offset + 3L * exp_dim)] - block_ep)),
                1e-10)
      offset <- offset + 3L * exp_dim

      ## exp_dual block (everything that's left)
      block_ed <- diffcp::pi(x[(offset + 1L):length(x)],
                             list(list(name = "ed", sizes = exp_dim)),
                             dual = dual)
      expect_lt(max(abs(proj[(offset + 1L):length(x)] - block_ed)), 1e-10)
    }
  }
})

# === test_dpi (multi-cone Jacobian composition) ==================
test_that("D pi(x) ~= (pi(x+dx) - pi(x)) / FD on multi-cone (10 random)", {
  set.seed(0)
  for (rep in 1:10) {
    zero_dim <- sample.int(9L, 1L)
    pos_dim  <- sample.int(9L, 1L)
    soc_dim  <- sample.int(9L, sample.int(9L, 1L), replace = TRUE)
    psd_dim  <- sample.int(9L, sample.int(9L, 1L), replace = TRUE)
    exp_dim  <- sample.int(15L, 1L) + 2L

    cones <- list(
      list(name = "z",  sizes = zero_dim),
      list(name = "l",  sizes = pos_dim),
      list(name = "q",  sizes = as.integer(soc_dim)),
      list(name = "s",  sizes = as.integer(psd_dim)),
      list(name = "ep", sizes = exp_dim),
      list(name = "ed", sizes = exp_dim)
    )
    size <- zero_dim + pos_dim + sum(soc_dim) +
      sum(sapply(psd_dim, function(d) (d * (d + 1L)) %/% 2L)) +
      2L * 3L * exp_dim
    x <- rnorm(size)

    for (dual in c(FALSE, TRUE)) {
      proj_x <- diffcp::pi(x, cones, dual = dual)
      dx <- 1e-7 * rnorm(size)
      z <- diffcp::pi(x + dx, cones, dual = dual)
      Dpi_dense <- diffcp:::cpp_dprojection_dense(x, cones, dual)
      expect_lt(max(abs(Dpi_dense %*% dx - (z - proj_x))), 1e-6)
    }
  }
})

# === test_lsqr ===================================================
test_that("LSQR on random A, b matches base lstsq (qr.solve)", {
  set.seed(0)
  A <- matrix(rnorm(20 * 10), 20, 10)
  b <- rnorm(20)
  A_csc <- methods::as(Matrix::Matrix(A, sparse = TRUE), "CsparseMatrix")
  out <- diffcp:::cpp_lsqr_sparse(A_csc, b)
  svx <- qr.solve(A, b)
  expect_lt(max(abs(svx - out$solution)), 1e-8)
})

# === test_infeasible =============================================
test_that("solve_only errors on a primal-infeasible 1D LP (SCS)", {
  ## c = [1], A = [[1], [1]], b = [1, -1] -> x must be both 1 and -1.
  ## Python uses cone_dims = {"f": 2} (free / equality cone).
  A <- methods::as(Matrix::Matrix(matrix(c(1, 1), nrow = 2L), sparse = TRUE),
                   "CsparseMatrix")
  b <- c(1.0, -1.0)
  c <- c(1.0)
  expect_error(
    diffcp::solve_only(A, b, c, list(z = 2L), solve_method = "SCS"),
    "SCS returned"
  )
})

# === test_dprojection_exp (analytical sensitivity) ===============
test_that("D matches analytical sensitivity for entropy-parametric problem", {
  ## CVXPY problem (one variable + parameter lam):
  ##    maximize x + lam*(log(1+x) + log(1-x))
  ## At lam = 1, sensitivity dx/dlam = -1 + lam/sqrt(lam^2 + 1) = -1 + 1/sqrt(2).
  A <- .csc(dprojexp_A_A_i, dprojexp_A_A_p, dprojexp_A_A_x, dprojexp_A_shape)

  res <- diffcp::solve_and_derivative(
    A, dprojexp_b, dprojexp_c, dprojexp_cone_dict,
    solve_method = "Clarabel", mode = "dense",
    tol_gap_abs = 1e-13, tol_gap_rel = 1e-13,
    tol_feas = 1e-13, tol_ktratio = 1e-13)

  dlam <- 1e-6
  dA <- Matrix::sparseMatrix(i = integer(0), j = integer(0), x = numeric(0),
                             dims = dim(A))
  db <- numeric(length(dprojexp_b))
  dc <- numeric(length(dprojexp_c))
  ## Python comment (cone_prog_diff.py:388-389): the minus sign stems
  ## from the fact that c = [-1, -1, -1].
  dc[2] <- -dlam
  dc[3] <- -dlam
  d <- res$D(dA, db, dc)

  expect_lt(abs(dprojexp_analytical - d$dx[1] / dlam), 1e-6)
})

## R-SPECIFIC: upstream pin metadata helpers.
##
## The R port tracks an exact upstream Python `diffcp` commit. The
## canonical pin lives in `inst/UPSTREAM.dcf` (DCF so it's readable
## from R via `read.dcf`). Per-file `## DIFFCP SOURCE:` headers point
## at paths inside that pinned upstream tree.

#' Upstream Python diffcp pin
#'
#' Returns the upstream pin metadata for this R port: which
#' `cvxgrp/diffcp` version / commit the R sources track, the date of
#' that commit, and diffcp's own CVXPY runtime constraint. The
#' authoritative record lives in `inst/UPSTREAM.dcf` and is read at
#' call time.
#'
#' @return A named character vector with one element per DCF field
#'   (`Upstream`, `Version`, `Commit`, `ShortCommit`, `Date`, `URL`,
#'   `CVXPY`, `Snapshot`, `Notes`).
#' @export
#' @examples
#' upstream_info()
#' upstream_info()[["Version"]]
upstream_info <- function() {
  path <- system.file("UPSTREAM.dcf", package = "diffcp")
  if (!nzchar(path)) {
    stop("UPSTREAM.dcf not found in installed diffcp package.")
  }
  fields <- read.dcf(path)
  if (nrow(fields) != 1L) {
    stop("UPSTREAM.dcf must contain exactly one record.")
  }
  drop(fields)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

cpp_dprojection_dense <- function(x, cones_R, dual) {
    .Call(`_diffcp_cpp_dprojection_dense`, x, cones_R, dual)
}

cpp_M_dense <- function(Q_dense, cones_R, u, v, w) {
    .Call(`_diffcp_cpp_M_dense`, Q_dense, cones_R, u, v, w)
}

cpp_solve_derivative_dense <- function(M, MT, rhs) {
    .Call(`_diffcp_cpp_solve_derivative_dense`, M, MT, rhs)
}

cpp_solve_adjoint_derivative_dense <- function(M, MT, dz) {
    .Call(`_diffcp_cpp_solve_adjoint_derivative_dense`, M, MT, dz)
}

cpp_lsqr_M <- function(Q, cones_R, u, v, w, rhs, transpose, damp = 0.0, atol = 1e-8, btol = 1e-8, conlim = 1e8, iter_lim = -1L) {
    .Call(`_diffcp_cpp_lsqr_M`, Q, cones_R, u, v, w, rhs, transpose, damp, atol, btol, conlim, iter_lim)
}

cpp_lsqr_sparse <- function(A, b, damp = 0.0, atol = 1e-8, btol = 1e-8, conlim = 1e8, iter_lim = -1L) {
    .Call(`_diffcp_cpp_lsqr_sparse`, A, b, damp, atol, btol, conlim, iter_lim)
}

cpp_project_exp_cone <- function(x) {
    .Call(`_diffcp_cpp_project_exp_cone`, x)
}

cpp_in_exp <- function(x) {
    .Call(`_diffcp_cpp_in_exp`, x)
}

cpp_in_exp_dual <- function(x) {
    .Call(`_diffcp_cpp_in_exp_dual`, x)
}

## DIFFCP SOURCE: diffcp/cone_program.py @ v1.1.8 (6a78143)
## Upstream pin: see inst/UPSTREAM.dcf / upstream_info().
##
## Forward solve dispatch (Clarabel / SCS) plus the public
## `solve_only` and `solve_and_derivative` entry points. Derivative
## computation (M operator, LSQR, dense modes) lands in subsequent
## commits; for now `solve_and_derivative` errors with a clear
## phase-pointer message, while `solve_only` is fully functional.

# -- PSD permutation: SCS lower-triangular <-> Clarabel upper-triangular --
##
## DIFFCP SOURCE: diffcp/cone_program.py:14-82 (`permute_psd_rows`,
## `inverse_permute_psd_solution`).
##
## SCS packs a vec(X) for an n x n PSD matrix as the lower triangle in
## column-major order: (X11, X21, ..., Xn1, X22, X32, ..., Xnn).
## Clarabel packs the upper triangle in column-major order:
## (X11, X12, X22, X13, X23, X33, ...).  For symmetric X both encode
## the same matrix, but the *positions* differ, so the rows of A and
## entries of b corresponding to one PSD block must be permuted before
## handing to Clarabel; y and s come back in Clarabel order and must
## be permuted back to SCS order.
##
## We compute two permutations matching Python:
##   post[k] = clarabel index for the entry whose SCS index is k
##   pre[k]  = scs index for the entry whose clarabel index is k
## (post and pre are inverses of each other).
.psd_perms <- function(n) {
  ## np.tril_indices(n): for r=0..n-1, for c=0..r yield (r, c)
  ## np.triu_indices(n): for r=0..n-1, for c=r..n-1 yield (r, c)
  tril <- do.call(rbind, lapply(0:(n - 1L),
                                function(r) cbind(r, 0:r)))
  triu <- do.call(rbind, lapply(0:(n - 1L),
                                function(r) cbind(r, r:(n - 1L))))
  ## ravel_multi_index((cols, rows), (n, n)) row-major:
  ##   linear = cols * n + rows
  tril_multi <- tril[, 2L] * n + tril[, 1L]
  triu_multi <- triu[, 2L] * n + triu[, 1L]
  ## np.argsort -> R order(); convert to 0-based offsets.
  list(
    post = order(tril_multi) - 1L,   # SCS k -> Clarabel position
    pre  = order(triu_multi) - 1L    # Clarabel k -> SCS position
  )
}

## Permute rows of a CsparseMatrix A and entries of b for ONE PSD
## block of dimension `n` (n*(n+1)/2 vectorized entries) starting at
## 1-based row index `row_offset`.
.permute_psd_rows <- function(A, b, n, row_offset) {
  perms <- .psd_perms(n)
  n_rows <- length(perms$post)

  At <- methods::as(A, "TsparseMatrix")
  rows <- At@i + 1L                    # 1-based
  cols <- At@j + 1L
  vals <- At@x
  in_block <- rows >= row_offset & rows < row_offset + n_rows
  rows[in_block] <- row_offset + perms$post[rows[in_block] - row_offset + 1L]

  new_A <- Matrix::sparseMatrix(i = rows, j = cols, x = vals, dims = dim(A))
  new_b <- b
  new_b[row_offset:(row_offset + n_rows - 1L)] <-
    b[row_offset + perms$pre + 1L - 1L]   # 1-based
  list(A = new_A, b = new_b)
}

## Inverse permutation applied to y and s after a Clarabel solve.
.inverse_permute_psd_solution <- function(y, s, n, row_offset) {
  perms <- .psd_perms(n)
  n_rows <- length(perms$pre)
  idx <- row_offset + perms$pre   # 1-based source positions
  new_y <- y
  new_s <- s
  new_y[row_offset:(row_offset + n_rows - 1L)] <- y[idx]
  new_s[row_offset:(row_offset + n_rows - 1L)] <- s[idx]
  list(y = new_y, s = new_s)
}

# -- Internal: dispatch the forward solve to a concrete solver ----
## DIFFCP SOURCE: diffcp/cone_program.py:397-630 (solve_internal).
##
## Differences from the Python source in this development version:
##   * ECOS branch is omitted (R `ECOSolveR` API differs; deferred
##     until we need ECOS-specific behavior for tests).
##   * Clarabel PSD row-permutation (`permute_psd_rows` /
##     `inverse_permute_psd_solution`) is omitted — PSD support
##     arrives in Phase 2b alongside the PSD cone projection.
##   * SCS `acceleration_lookback=0` retry on inaccurate status is
##     dropped (R `scs` package takes the same control name and we
##     can add the retry once we hit an inaccurate case in tests).
.solve_internal <- function(A, b, c, cone_dict,
                            solve_method = NULL,
                            warm_start = NULL,
                            raise_on_error = TRUE,
                            P = NULL,
                            ...) {
  kwargs <- list(...)

  if (is.null(solve_method)) {
    psd_cone <- !is.null(cone_dict[["s"]]) && length(cone_dict[["s"]]) > 0L
    ed_cone  <- !is.null(cone_dict[["ed"]]) && cone_dict[["ed"]] > 0L
    solve_method <- if (psd_cone || ed_cone) "SCS" else "Clarabel"
  }

  if (solve_method == "Clarabel" || solve_method == "CLARABEL") {
    if (!requireNamespace("clarabel", quietly = TRUE)) {
      cli::cli_abort("The {.pkg clarabel} package is required for Clarabel.")
    }
    if (!is.null(warm_start)) {
      cli::cli_abort("Clarabel does not support warm starting.")
    }

    ## Clarabel uses the same SCS-style cones list. Translate `f`->`z`
    ## (free cone alias) and drop `ed` (Clarabel does not implement
    ## dual exponential cones natively).  PSD cones need an additional
    ## row permutation on A and b, applied below per cone in cone-order
    ## start_row = z + f + l + sum(q); then for each PSD of size v,
    ## permute and advance by v*(v+1)/2.
    cones_arg <- list()
    if (!is.null(cone_dict[["z"]]) && cone_dict[["z"]] > 0L)
      cones_arg$z <- as.integer(cone_dict[["z"]])
    if (!is.null(cone_dict[["f"]]) && cone_dict[["f"]] > 0L) {
      cones_arg$z <- (cones_arg$z %||% 0L) + as.integer(cone_dict[["f"]])
    }
    if (!is.null(cone_dict[["l"]]) && cone_dict[["l"]] > 0L)
      cones_arg$l <- as.integer(cone_dict[["l"]])
    if (!is.null(cone_dict[["q"]]) && length(cone_dict[["q"]]) > 0L)
      cones_arg$q <- as.integer(cone_dict[["q"]])
    if (!is.null(cone_dict[["s"]]) && length(cone_dict[["s"]]) > 0L)
      cones_arg$s <- as.integer(cone_dict[["s"]])
    if (!is.null(cone_dict[["ep"]]) && cone_dict[["ep"]] > 0L)
      cones_arg$ep <- as.integer(cone_dict[["ep"]])

    ## Apply PSD row permutations to A, b in SCS->Clarabel direction.
    if (!is.null(cone_dict[["s"]]) && length(cone_dict[["s"]]) > 0L) {
      start_row <- 1L  # 1-based
      if (!is.null(cone_dict[["z"]])) start_row <- start_row + as.integer(cone_dict[["z"]])
      if (!is.null(cone_dict[["f"]])) start_row <- start_row + as.integer(cone_dict[["f"]])
      if (!is.null(cone_dict[["l"]])) start_row <- start_row + as.integer(cone_dict[["l"]])
      if (!is.null(cone_dict[["q"]])) start_row <- start_row + sum(as.integer(cone_dict[["q"]]))
      for (v in as.integer(cone_dict[["s"]])) {
        permuted <- .permute_psd_rows(A, b, v, start_row)
        A <- permuted$A
        b <- permuted$b
        start_row <- start_row + (v * (v + 1L)) %/% 2L
      }
    }

    if (is.null(P)) {
      n <- length(c)
      P <- Matrix::sparseMatrix(i = integer(0), j = integer(0), x = numeric(0),
                                dims = c(n, n))
    }

    control <- do.call(clarabel::clarabel_control, c(list(verbose = FALSE), kwargs))
    A_csc <- methods::as(A, "CsparseMatrix")
    P_csc <- methods::as(P, "CsparseMatrix")
    sol <- clarabel::clarabel(A = A_csc, b = b, q = c, P = P_csc,
                              cones = cones_arg, control = control)

    ## R clarabel: `sol$status` is an integer index into
    ## `solver_status_descriptions()`; the *names* of that vector are
    ## the canonical status tags ("Solved", "PrimalInfeasible", ...)
    ## that match Python clarabel's `str(solution.status)`.
    status_str  <- names(clarabel::solver_status_descriptions())[sol$status]
    CLARABEL2SCS_STATUS_MAP <- c(
      "Solved"                  = "Solved",
      "PrimalInfeasible"        = "Infeasible",
      "DualInfeasible"          = "Unbounded",
      "AlmostSolved"            = "Optimal Inaccurate",
      "AlmostPrimalInfeasible"  = "Infeasible Inaccurate",
      "AlmostDualInfeasible"    = "Unbounded Inaccurate"
    )
    scs_status <- CLARABEL2SCS_STATUS_MAP[status_str]
    if (is.na(scs_status)) scs_status <- "Failure"

    inaccurate_status <- c("Optimal Inaccurate", "Infeasible Inaccurate",
                           "Unbounded Inaccurate")
    if (scs_status == "Optimal Inaccurate") {
      ## Python's Clarabel branch (cone_program.py:540-626) does NOT
      ## error on AlmostSolved; downstream code accepts the answer.
      ## We only warn here.
      cli::cli_warn("Clarabel returned {.val {status_str}}.")
    } else if (scs_status != "Solved") {
      if (raise_on_error) {
        cli::cli_abort("Clarabel returned status {.val {status_str}}.")
      } else {
        return(list(x = sol$x, y = sol$z, s = sol$s,
                    info = list(status = scs_status), D = NULL, DT = NULL))
      }
    }

    y_out <- as.numeric(sol$z)
    s_out <- as.numeric(sol$s)

    ## Permute y and s back from Clarabel (upper-triangular) to SCS
    ## (lower-triangular) for each PSD cone block.
    if (!is.null(cone_dict[["s"]]) && length(cone_dict[["s"]]) > 0L) {
      start_row <- 1L
      if (!is.null(cone_dict[["z"]])) start_row <- start_row + as.integer(cone_dict[["z"]])
      if (!is.null(cone_dict[["f"]])) start_row <- start_row + as.integer(cone_dict[["f"]])
      if (!is.null(cone_dict[["l"]])) start_row <- start_row + as.integer(cone_dict[["l"]])
      if (!is.null(cone_dict[["q"]])) start_row <- start_row + sum(as.integer(cone_dict[["q"]]))
      for (v in as.integer(cone_dict[["s"]])) {
        permuted <- .inverse_permute_psd_solution(y_out, s_out, v, start_row)
        y_out <- permuted$y
        s_out <- permuted$s
        start_row <- start_row + (v * (v + 1L)) %/% 2L
      }
    }

    return(list(
      x = as.numeric(sol$x),
      y = y_out,
      s = s_out,
      info = list(
        status    = scs_status,
        solveTime = sol$solve_time %||% NA_real_,
        setupTime = -1,
        iter      = sol$iterations %||% NA_integer_,
        pobj      = sol$obj_val    %||% NA_real_
      )
    ))
  }

  if (solve_method == "SCS") {
    if (!requireNamespace("scs", quietly = TRUE)) {
      cli::cli_abort("The {.pkg scs} package is required for SCS.")
    }

    ## R `scs::scs` uses `eps_abs` / `eps_rel` (matching SCS >= 3.x).
    ## Mirror the Python convenience: a single `eps` becomes both.
    if (!is.null(kwargs[["eps"]])) {
      kwargs[["eps_abs"]] <- kwargs[["eps"]]
      kwargs[["eps_rel"]] <- kwargs[["eps"]]
      kwargs[["eps"]] <- NULL
    }
    kwargs$verbose <- kwargs$verbose %||% FALSE
    control <- do.call(scs::scs_control, kwargs)

    initial <- NULL
    if (!is.null(warm_start)) {
      initial <- list(x = warm_start[[1L]], y = warm_start[[2L]], s = warm_start[[3L]])
    }

    A_csc <- methods::as(A, "CsparseMatrix")
    ## R `scs` accepts P only as base `matrix` or
    ## `slam::simple_triplet_matrix` (no Matrix-package CsparseMatrix
    ## method).  Densify if sparse — fine in practice because P is
    ## typically small.
    P_in <- if (is.null(P)) NULL else as.matrix(P)
    sol <- scs::scs(A = A_csc, b = b, obj = c, P = P_in,
                    cone = cone_dict, initial = initial, control = control)

    status <- sol$info$status
    inaccurate_status <- c("Solved/Inaccurate",
                           "solved (inaccurate - reached max_iters)",
                           "solved (inaccurate - reached time_limit_secs)")
    if (status %in% inaccurate_status) {
      cli::cli_warn("Solved/Inaccurate.")
    } else if (tolower(status) != "solved") {
      if (raise_on_error) {
        cli::cli_abort("SCS returned status {.val {status}}.")
      } else {
        return(list(x = sol$x, y = sol$y, s = sol$s,
                    info = sol$info, D = NULL, DT = NULL))
      }
    }

    return(list(
      x = as.numeric(sol$x),
      y = as.numeric(sol$y),
      s = as.numeric(sol$s),
      info = sol$info
    ))
  }

  cli::cli_abort("Solver {.val {solve_method}} not supported.")
}

`%||%` <- function(a, b) if (is.null(a)) b else a

#' Solve a cone program (forward only)
#'
#' Mirrors `diffcp.cone_program.solve_only`. Solves
#'   minimize  c^T x  (+ 0.5 x^T P x)
#'   subject to  A x + s = b,   s in K
#' and returns the optimal `(x, y, s)`. Unlike `solve_and_derivative`
#' this function does not build the derivative closures.
#'
#' @param A A sparse `dgCMatrix` constraint matrix.
#' @param b A numeric offset vector.
#' @param c A numeric objective coefficient vector.
#' @param cone_dict A named list with cone sizes (keys among
#'   `"z"`, `"l"`, `"q"`, `"s"`, `"ep"`, `"ed"`).
#' @param warm_start Optional warm-start `list(x, y, s)`.
#' @param solve_method One of `"Clarabel"` (default) or `"SCS"`.
#' @param P Optional sparse `dgCMatrix` for QP objective.
#' @param ... Additional control parameters forwarded to the solver.
#' @returns A list with elements `x`, `y`, `s`, `info`.
#' @export
solve_only <- function(A, b, c, cone_dict,
                       warm_start = NULL,
                       solve_method = "Clarabel",
                       P = NULL,
                       ...) {
  if (any(is.nan(A@x))) cli::cli_abort("Found a NaN in {.arg A}.")
  ## Eliminate explicit zeros (matches Python `A.eliminate_zeros()`).
  A <- Matrix::drop0(A)
  res <- .solve_internal(A, b, c, cone_dict,
                         warm_start = warm_start,
                         solve_method = solve_method,
                         P = P, ...)
  list(x = res$x, y = res$y, s = res$s, info = res$info)
}

#' Solve a cone program and return forward / adjoint derivative operators
#'
#' Mirrors `diffcp.cone_program.solve_and_derivative`. Solves
#'   minimize  c^T x        s.t.  A x + s = b,  s in K
#' (with optional QP `0.5 x^T P x` term) and returns the optimal
#' `(x, y, s)` together with closures `D` (forward) and `DT` (adjoint)
#' that map perturbations of `(A, b, c, [P])` to perturbations of
#' `(x, y, s)` and vice versa.
#'
#' @inheritParams solve_only
#' @param mode Differentiation mode: `"lsqr"` (default), `"dense"`,
#'   or `"lpgd"`.
#' @returns A list with elements `x`, `y`, `s`, `info`, `D`, `DT`.
#'   `info` is the solver-status block returned by the underlying
#'   solver (`status`, `iter`, `solveTime`, `pobj`, ...); see
#'   `solve_only` for the same shape.
#' @export
solve_and_derivative <- function(A, b, c, cone_dict,
                                 P = NULL,
                                 solve_method = "Clarabel",
                                 mode = "lsqr",
                                 warm_start = NULL,
                                 ...) {
  if (!mode %in% c("dense", "lsqr")) {
    cli::cli_abort("Unsupported {.arg mode} = {.val {mode}}; supported: 'dense', 'lsqr'.")
  }
  if (!is.null(P)) {
    ## Mirror Python diffcp/cone_program.py:637-639.  Python diffcp
    ## supports QP only via lpgd modes (lpgd / lpgd_left / lpgd_right);
    ## we have not ported lpgd, so the path errors with the same text.
    cli::cli_abort(c(
      "Dense, lsqr, and lsmr modes currently do not support quadratic objectives.",
      "i" = "Consider switching to {.code mode = 'lpgd'} mode.",
      "x" = "{.code mode = 'lpgd'} is not yet ported in this R version of diffcp."
    ))
  }
  if (any(is.nan(A@x))) cli::cli_abort("Found a NaN in {.arg A}.")

  ## Capture the sparsity pattern of `A` INCLUDING any explicit zeros, BEFORE
  ## eliminating them.  `dA` is reported on this pattern, so an entry that is
  ## structurally present but numerically zero still gets a derivative.
  ##
  ## DIFFCP SOURCE: cone_program.py:643-653, which is explicit about it:
  ##     A.data[A.data == 0] = np.nan   # op1: mark the explicit zeros
  ##     rows, cols = A.nonzero()       # op2: capture the pattern WITH them
  ##     A.data[np.isnan(A.data)] = 0.0 # op3: restore
  ##     A.eliminate_zeros()            # the solver gets a clean A
  ## and then builds `dA` on that captured `(rows, cols)` at :771 / :891.
  ##
  ## R needs no op1/op3: `Matrix::summary()` on a dgCMatrix lists STORED
  ## entries, explicit zeros included, so the pattern can be read directly.
  ## Do NOT "restore parity" by adding the NaN marking back -- it would be a
  ## no-op here at best, and `A@x` is shared structure.
  ##
  ## Why this matters: a caller that differentiates with respect to a
  ## parameter driving an entry of `A` stores that entry EXPLICITLY, even when
  ## its current value is 0 (CVXPY calls this `keep_zeros`, and it is the sole
  ## reason its DIFFCP interface overrides `apply`).  Dropping the zero first
  ## removed the entry from `dA` altogether and the parameter's gradient came
  ## back as exactly 0 -- silently, with no error and no warning.
  A_pattern <- Matrix::summary(methods::as(A, "CsparseMatrix"))
  A <- Matrix::drop0(A)

  res <- .solve_internal(A, b, c, cone_dict,
                         warm_start = warm_start,
                         solve_method = solve_method,
                         P = P, ...)
  x <- res$x; y <- res$y; s <- res$s

  cones <- parse_cone_dict(cone_dict)
  closures <- .make_derivative_closures(A, b, c, x, y, s, cones, mode,
                                        pattern = A_pattern)

  list(x = x, y = y, s = s, info = res$info,
       D = closures$D, DT = closures$DT)
}

## Faithful R ports of the canonical Python diffcp tests:
##
##   diffcp/tests/test_clarabel.py::test_solve_and_derivative
##   diffcp/tests/test_clarabel.py::test_psdcone
##   diffcp/tests/test_scs.py::test_solve_and_derivative
##   diffcp/tests/test_scs.py::test_warm_start
##   diffcp/tests/test_scs.py::test_expcone
##
## (A, b, c, cone_dict) and the deterministic random perturbations
## come from `inst/python_parity/canonical.py`, dumped into
## `helper-canonical-data.R` (auto-sourced by testthat).
##
## Same atol bounds as Python: 1e-8 on D() finite-diff agreement and
## adjoint objective sensitivity; 1e-7 on warm-start (x_p ~= x).

skip_if_not_installed("clarabel")
skip_if_not_installed("scs")

.csc <- function(A_i, A_p, A_x, dims) {
  Matrix::sparseMatrix(i = A_i, p = A_p - 1L, x = A_x, dims = dims, index1 = TRUE)
}

# ---- helper: build dA from same nonzero pattern, given pinned values --
.dA_from_pattern <- function(A_i, A_p, dA_data, dims) {
  Matrix::sparseMatrix(i = A_i, p = A_p - 1L, x = dA_data, dims = dims, index1 = TRUE)
}

# === Port: test_clarabel.py::test_solve_and_derivative =================
test_that("clarabel: D matches finite-diff on LS-eq problem (mode='lsqr')", {
  A <- .csc(clarabel_lseq_A_A_i, clarabel_lseq_A_A_p,
            clarabel_lseq_A_A_x, clarabel_lseq_A_shape)
  res <- diffcp::solve_and_derivative(A, clarabel_lseq_b, clarabel_lseq_c,
                                      clarabel_lseq_cone_dict,
                                      mode = "lsqr",
                                      solve_method = "Clarabel")
  dA <- .dA_from_pattern(clarabel_lseq_A_A_i, clarabel_lseq_A_A_p,
                         clarabel_lseq_lsqr_dA_data, clarabel_lseq_A_shape)
  d <- res$D(dA, clarabel_lseq_lsqr_db, clarabel_lseq_lsqr_dc)

  res_pert <- diffcp::solve_only(A + dA,
                                 clarabel_lseq_b + clarabel_lseq_lsqr_db,
                                 clarabel_lseq_c + clarabel_lseq_lsqr_dc,
                                 clarabel_lseq_cone_dict,
                                 solve_method = "Clarabel")

  expect_lt(max(abs(d$dx - (res_pert$x - res$x))), 1e-8)
  expect_lt(max(abs(d$dy - (res_pert$y - res$y))), 1e-8)
  expect_lt(max(abs(d$ds - (res_pert$s - res$s))), 1e-8)
})

test_that("clarabel: D matches finite-diff on LS-eq problem (mode='dense')", {
  A <- .csc(clarabel_lseq_A_A_i, clarabel_lseq_A_A_p,
            clarabel_lseq_A_A_x, clarabel_lseq_A_shape)
  res <- diffcp::solve_and_derivative(A, clarabel_lseq_b, clarabel_lseq_c,
                                      clarabel_lseq_cone_dict,
                                      mode = "dense",
                                      solve_method = "Clarabel")
  dA <- .dA_from_pattern(clarabel_lseq_A_A_i, clarabel_lseq_A_A_p,
                         clarabel_lseq_dense_dA_data, clarabel_lseq_A_shape)
  d <- res$D(dA, clarabel_lseq_dense_db, clarabel_lseq_dense_dc)

  res_pert <- diffcp::solve_only(A + dA,
                                 clarabel_lseq_b + clarabel_lseq_dense_db,
                                 clarabel_lseq_c + clarabel_lseq_dense_dc,
                                 clarabel_lseq_cone_dict,
                                 solve_method = "Clarabel")

  expect_lt(max(abs(d$dx - (res_pert$x - res$x))), 1e-8)
  expect_lt(max(abs(d$dy - (res_pert$y - res$y))), 1e-8)
  expect_lt(max(abs(d$ds - (res_pert$s - res$s))), 1e-8)
})

# === Port: test_clarabel.py::test_psdcone =============================
test_that("clarabel: 5x5 min-eig SDP forward solve (test_psdcone)", {
  A <- .csc(clarabel_psdcone_A_A_i, clarabel_psdcone_A_A_p,
            clarabel_psdcone_A_A_x, clarabel_psdcone_A_shape)
  res <- diffcp::solve_and_derivative(A, clarabel_psdcone_b, clarabel_psdcone_c,
                                      clarabel_psdcone_cone_dict,
                                      solve_method = "Clarabel")
  ## Python asserts trace(unvec_symm(sol_vec)) ~= 1 at 1e-6 and all
  ## eigenvalues >= -1e-6.
  sol <- diffcp::unvec_symm(res$x, 5L)
  expect_lt(abs(sum(diag(sol)) - 1), 1e-6)
  evs <- eigen(sol, symmetric = TRUE, only.values = TRUE)$values
  expect_gt(min(evs), -1e-6)
})

# === Port: test_scs.py::test_solve_and_derivative =====================
test_that("scs: D matches finite-diff on LS-eq problem (mode='lsqr')", {
  A <- .csc(scs_lseq_A_A_i, scs_lseq_A_A_p,
            scs_lseq_A_A_x, scs_lseq_A_shape)
  res <- diffcp::solve_and_derivative(A, scs_lseq_b, scs_lseq_c,
                                      scs_lseq_cone_dict,
                                      mode = "lsqr",
                                      solve_method = "SCS",
                                      eps = 1e-10,
                                      adaptive_scale = FALSE)
  dA <- .dA_from_pattern(scs_lseq_A_A_i, scs_lseq_A_A_p,
                         scs_lseq_lsqr_dA_data, scs_lseq_A_shape)
  d <- res$D(dA, scs_lseq_lsqr_db, scs_lseq_lsqr_dc)

  res_pert <- diffcp::solve_only(A + dA,
                                 scs_lseq_b + scs_lseq_lsqr_db,
                                 scs_lseq_c + scs_lseq_lsqr_dc,
                                 scs_lseq_cone_dict,
                                 solve_method = "SCS",
                                 eps = 1e-10,
                                 adaptive_scale = FALSE)

  expect_lt(max(abs(d$dx - (res_pert$x - res$x))), 1e-8)
  expect_lt(max(abs(d$dy - (res_pert$y - res$y))), 1e-8)
  expect_lt(max(abs(d$ds - (res_pert$s - res$s))), 1e-8)
})

test_that("scs: D matches finite-diff on LS-eq problem (mode='dense')", {
  A <- .csc(scs_lseq_A_A_i, scs_lseq_A_A_p,
            scs_lseq_A_A_x, scs_lseq_A_shape)
  res <- diffcp::solve_and_derivative(A, scs_lseq_b, scs_lseq_c,
                                      scs_lseq_cone_dict,
                                      mode = "dense",
                                      solve_method = "SCS",
                                      eps = 1e-10,
                                      adaptive_scale = FALSE)
  dA <- .dA_from_pattern(scs_lseq_A_A_i, scs_lseq_A_A_p,
                         scs_lseq_dense_dA_data, scs_lseq_A_shape)
  d <- res$D(dA, scs_lseq_dense_db, scs_lseq_dense_dc)

  res_pert <- diffcp::solve_only(A + dA,
                                 scs_lseq_b + scs_lseq_dense_db,
                                 scs_lseq_c + scs_lseq_dense_dc,
                                 scs_lseq_cone_dict,
                                 solve_method = "SCS",
                                 eps = 1e-10,
                                 adaptive_scale = FALSE)

  expect_lt(max(abs(d$dx - (res_pert$x - res$x))), 1e-8)
  expect_lt(max(abs(d$dy - (res_pert$y - res$y))), 1e-8)
  expect_lt(max(abs(d$ds - (res_pert$s - res$s))), 1e-8)
})

# === Port: test_scs.py::test_warm_start ===============================
test_that("scs: warm-start solve agrees with cold solve (max_iters=1)", {
  A <- .csc(scs_warm_start_A_A_i, scs_warm_start_A_A_p,
            scs_warm_start_A_A_x, scs_warm_start_A_shape)
  res <- diffcp::solve_and_derivative(A, scs_warm_start_b, scs_warm_start_c,
                                      scs_warm_start_cone_dict,
                                      solve_method = "SCS", eps = 1e-9)
  ## With warm-start at the optimum and max_iters = 1, SCS should
  ## return essentially the same point.
  res_p <- diffcp::solve_and_derivative(A, scs_warm_start_b, scs_warm_start_c,
                                        scs_warm_start_cone_dict,
                                        warm_start = list(res$x, res$y, res$s),
                                        max_iters = 1L,
                                        solve_method = "SCS", eps = 1e-9)
  expect_lt(max(abs(res$x - res_p$x)), 1e-7)
  expect_lt(max(abs(res$y - res_p$y)), 1e-7)
  expect_lt(max(abs(res$s - res_p$s)), 1e-7)
})

# === Port: test_scs.py::test_expcone ==================================
test_that("scs: D matches finite-diff on entropy-max EXP cone (mode='lsqr')", {
  A <- .csc(scs_expcone_A_A_i, scs_expcone_A_A_p,
            scs_expcone_A_A_x, scs_expcone_A_shape)
  res <- diffcp::solve_and_derivative(A, scs_expcone_b, scs_expcone_c,
                                      scs_expcone_cone_dict,
                                      mode = "lsqr", solve_method = "SCS",
                                      eps = 1e-10)
  dA <- .dA_from_pattern(scs_expcone_A_A_i, scs_expcone_A_A_p,
                         scs_expcone_lsqr_dA_data, scs_expcone_A_shape)
  d <- res$D(dA, scs_expcone_lsqr_db, scs_expcone_lsqr_dc)

  res_pert <- diffcp::solve_only(A + dA,
                                 scs_expcone_b + scs_expcone_lsqr_db,
                                 scs_expcone_c + scs_expcone_lsqr_dc,
                                 scs_expcone_cone_dict,
                                 solve_method = "SCS", eps = 1e-10)

  expect_lt(max(abs(d$dx - (res_pert$x - res$x))), 1e-8)
  expect_lt(max(abs(d$dy - (res_pert$y - res$y))), 1e-8)
  expect_lt(max(abs(d$ds - (res_pert$s - res$s))), 1e-8)
})

test_that("scs: D matches finite-diff on entropy-max EXP cone (mode='dense')", {
  A <- .csc(scs_expcone_A_A_i, scs_expcone_A_A_p,
            scs_expcone_A_A_x, scs_expcone_A_shape)
  res <- diffcp::solve_and_derivative(A, scs_expcone_b, scs_expcone_c,
                                      scs_expcone_cone_dict,
                                      mode = "dense", solve_method = "SCS",
                                      eps = 1e-10)
  dA <- .dA_from_pattern(scs_expcone_A_A_i, scs_expcone_A_A_p,
                         scs_expcone_dense_dA_data, scs_expcone_A_shape)
  d <- res$D(dA, scs_expcone_dense_db, scs_expcone_dense_dc)

  res_pert <- diffcp::solve_only(A + dA,
                                 scs_expcone_b + scs_expcone_dense_db,
                                 scs_expcone_c + scs_expcone_dense_dc,
                                 scs_expcone_cone_dict,
                                 solve_method = "SCS", eps = 1e-10)

  expect_lt(max(abs(d$dx - (res_pert$x - res$x))), 1e-8)
  expect_lt(max(abs(d$dy - (res_pert$y - res$y))), 1e-8)
  expect_lt(max(abs(d$ds - (res_pert$s - res$s))), 1e-8)
})

## Explicit zeros in `A` must survive into the pattern of `dA`.
##
## An entry of `A` that is STRUCTURALLY present but numerically zero is not the
## same thing as an absent entry: the first has a derivative, the second does
## not. Callers that differentiate with respect to a parameter driving that
## entry store it explicitly for exactly this reason -- CVXPY calls the flag
## `keep_zeros`, and it is the only reason its DIFFCP interface overrides
## `apply` (cvxpy/reductions/solvers/conic_solvers/diffcp_conif.py:72).
##
## Python diffcp preserves the pattern by marking explicit zeros as NaN,
## reading `A.nonzero()`, restoring them, and only then calling
## `eliminate_zeros()` (cone_program.py:643-653); `dA` is built on the captured
## `(rows, cols)` at :771 and :891. R needs no NaN dance -- `Matrix::summary()`
## already lists stored entries, explicit zeros included -- but the ORDER still
## matters, and this file exists because it was previously wrong: `drop0()` ran
## first and the pattern was read afterwards, so the entry vanished from `dA`
## and its gradient came back as exactly 0, with no error and no warning.

## minimize  c'x  s.t.  Ax + s = b, s >= 0, with an EXPLICIT ZERO at A[1,1].
## Feasible region x1 <= 3, x2 <= 5, and 0*x1 + x2 <= 1 -> x = (3, 1).
.ez_problem <- function(a11 = 0) {
  A <- Matrix::sparseMatrix(
    i = c(1L, 2L, 4L, 1L, 3L, 5L),
    j = c(1L, 1L, 1L, 2L, 2L, 2L),
    x = c(a11, -1,  1,  1, -1,  1),
    dims = c(5L, 2L))
  list(A = A, b = c(1, 0, 0, 3, 5), c = c(-1, -2), cone = list(z = 0L, l = 5L))
}

test_that("Matrix::summary() lists explicit zeros (the assumption behind the fix)", {
  m <- Matrix::sparseMatrix(i = c(1L, 2L), j = c(1L, 1L), x = c(0, -1),
                            dims = c(2L, 2L))
  expect_equal(nrow(Matrix::summary(m)), 2L)
  expect_equal(nrow(Matrix::summary(Matrix::drop0(m))), 1L)
})

test_that("an explicit zero in A keeps its entry in dA", {
  p <- .ez_problem(a11 = 0)
  expect_equal(length(p$A@x), 6L)         # the zero really is stored

  res <- solve_and_derivative(p$A, p$b, p$c, p$cone,
                              solve_method = "Clarabel", mode = "lsqr")
  d <- res$DT(rep(1, length(res$x)), numeric(length(res$y)), numeric(length(res$s)))
  nz <- Matrix::summary(d$dA)

  ## The pattern of dA is the pattern of A, explicit zeros included.
  expect_equal(nrow(nz), 6L)
  expect_true(any(nz$i == 1L & nz$j == 1L))

  ## And the derivative there is the real one, not a placeholder. Its value is
  ## d(sum(x))/d(A[1,1]); Python diffcp returns -3.00000001 on this problem.
  expect_equal(nz$x[nz$i == 1L & nz$j == 1L], -3, tolerance = 1e-6)
})

test_that("the derivative at an explicit zero agrees with a finite difference", {
  h <- 1e-6
  obj <- function(a11) {
    p <- .ez_problem(a11 = a11)
    r <- solve_only(Matrix::drop0(p$A), p$b, p$c, p$cone, solve_method = "Clarabel")
    sum(r$x)
  }
  fd <- (obj(h) - obj(-h)) / (2 * h)

  p <- .ez_problem(a11 = 0)
  res <- solve_and_derivative(p$A, p$b, p$c, p$cone,
                              solve_method = "Clarabel", mode = "lsqr")
  d <- res$DT(rep(1, length(res$x)), numeric(length(res$y)), numeric(length(res$s)))
  nz <- Matrix::summary(d$dA)
  analytic <- nz$x[nz$i == 1L & nz$j == 1L]

  expect_equal(analytic, fd, tolerance = 1e-4)
  ## The point of the whole exercise: it is NOT zero.
  expect_gt(abs(analytic), 1)
})

test_that("a matrix with no explicit zeros is completely unaffected", {
  ## drop0() is the identity here and summary() gives the same pattern, so the
  ## captured-pattern path must reproduce the old numbers exactly. This is what
  ## makes the change safe for every existing caller.
  p <- .ez_problem(a11 = 0)
  p$A <- Matrix::drop0(p$A)
  expect_equal(length(p$A@x), 5L)

  res <- solve_and_derivative(p$A, p$b, p$c, p$cone,
                              solve_method = "Clarabel", mode = "lsqr")
  d <- res$DT(rep(1, length(res$x)), numeric(length(res$y)), numeric(length(res$s)))
  expect_equal(nrow(Matrix::summary(d$dA)), 5L)
  expect_false(any(Matrix::summary(d$dA)$i == 1L & Matrix::summary(d$dA)$j == 1L))
})

test_that("dense mode preserves the pattern too", {
  p <- .ez_problem(a11 = 0)
  res <- solve_and_derivative(p$A, p$b, p$c, p$cone,
                              solve_method = "Clarabel", mode = "dense")
  d <- res$DT(rep(1, length(res$x)), numeric(length(res$y)), numeric(length(res$s)))
  nz <- Matrix::summary(d$dA)
  expect_equal(nrow(nz), 6L)
  expect_equal(nz$x[nz$i == 1L & nz$j == 1L], -3, tolerance = 1e-6)
})

test_that("solve_only still eliminates explicit zeros (no pattern to preserve)", {
  ## solve_only corresponds to Python cone_program.py:386, where no pattern is
  ## captured; its drop0 is correct and must stay.
  p <- .ez_problem(a11 = 0)
  r <- solve_only(p$A, p$b, p$c, p$cone, solve_method = "Clarabel")
  expect_equal(r$x, c(3, 1), tolerance = 1e-6)
})

## Forward QP solve via Clarabel and SCS, mirroring how cvxpylayers
## calls `solve_only(P=P)` in Python diffcp.
##
## Python diffcp's QP support is forward-only: solve_and_derivative
## with P != NULL and mode in {"dense", "lsqr", "lsmr"} explicitly
## errors and points the user to lpgd mode (cone_program.py:637-639).
## We mirror both behaviors here:
##   * solve_only(P) returns x/y/s matching Python's reference for
##     the same problem.
##   * solve_and_derivative(P) errors with Python's exact message
##     pointing to lpgd.
##
## Reference numerics from inst/python_parity/qp.py.

skip_if_not_installed("clarabel")
skip_if_not_installed("scs")

.csc <- function(A_i, A_p, A_x, dims) {
  Matrix::sparseMatrix(i = A_i, p = A_p - 1L, x = A_x, dims = dims, index1 = TRUE)
}

A <- .csc(c(1, 2, 1, 3, 1, 4), c(1, 3, 5, 7),
          c(1, -1, 1, -1, 1, -1), c(4, 3))
P <- .csc(c(1, 2, 1, 2, 3, 2, 3), c(1, 3, 6, 8),
          c(2, 0.5, 0.5, 2, 0.5, 0.5, 2), c(3, 3))
b <- c(1, 0, 0, 0)
c <- c(-1, 0.5, 1)
cone_dict <- list(z = 1L, l = 3L)

test_that("solve_only forward QP matches Python diffcp (Clarabel)", {
  res <- diffcp::solve_only(A, b, c, cone_dict, P = P,
                            solve_method = "Clarabel",
                            tol_gap_abs = 1e-12,
                            tol_gap_rel = 1e-12,
                            tol_feas    = 1e-12)
  expect_equal(res$x,
               c(0.9999997064006334, 1.590246952183832e-07,
                 1.345746671188647e-07),
               tolerance = 1e-7)
  expect_equal(res$y,
               c(-0.9999994923135322, 9.167811022181554e-14,
                  7.4622350325403e-07,  8.56348145174723e-07),
               tolerance = 1e-7)
  expect_equal(res$s,
               c(0, 0.9999997064006332, 1.590246889842621e-07,
                  1.345746599683592e-07),
               tolerance = 1e-7)
})

test_that("solve_only forward QP matches Python diffcp (SCS)", {
  res <- diffcp::solve_only(A, b, c, cone_dict, P = P,
                            solve_method = "SCS",
                            eps_abs = 1e-12, eps_rel = 1e-12,
                            max_iters = 100000L)
  expect_equal(res$x,
               c(1, 5.051524543201496e-17, -2.971459647020941e-17),
               tolerance = 1e-7)
  expect_equal(res$s,
               c(0, 1, 0, 0),
               tolerance = 1e-7)
})

test_that("solve_and_derivative aborts on QP with the upstream message", {
  ## Python diffcp/cone_program.py:637-639 raises:
  ##   "Dense, lsqr, and lsmr modes currently do not support quadratic
  ##    objectives. Consider switching to 'lpgd' mode."
  ## We mirror that exact text (lpgd is not yet ported in our R port).
  expect_error(
    diffcp::solve_and_derivative(A, b, c, cone_dict, P = P,
                                 mode = "dense"),
    "do not support quadratic objectives"
  )
  expect_error(
    diffcp::solve_and_derivative(A, b, c, cone_dict, P = P,
                                 mode = "lsqr"),
    "do not support quadratic objectives"
  )
})

## DIFFCP SOURCE: diffcp/__init__.py @ v1.1.8 (6a78143)
## Upstream pin: see inst/UPSTREAM.dcf / upstream_info().
##
## Package-level documentation and namespace setup.

#' diffcp: Differentiating Through Cone Programs
#'
#' R port of the Python `diffcp` package, mirroring Agrawal, Barratt,
#' Boyd, Busseti, and Moursi (2019).  Computes the derivative of the
#' optimal solution map of a convex cone program with respect to its
#' data, via the implicit-function theorem on the homogeneous self-dual
#' embedding.
#'
#' If you use this package in published work, please cite *both* the
#' R package and the original paper; see `citation("diffcp")` for the
#' full pair of BibTeX entries.
#'
#' @name diffcp-package
#' @aliases diffcp
#' @keywords internal
#' @import Matrix
#' @importFrom Rcpp evalCpp
#' @useDynLib diffcp, .registration = TRUE
"_PACKAGE"

## DIFFCP SOURCE: diffcp/cones.py:5-11

#' Cone tag constants
#'
#' String keys matching the SCS / Clarabel cone conventions. Used as
#' names in the `cone_dict` argument to [solve_only()] and
#' [solve_and_derivative()], and in the `(name, size)` pairs returned
#' by [parse_cone_dict()].
#'
#' @format Character scalars (and one character vector, `CONES`).
#' @details
#'   * `ZERO` (`"z"`): zero cone (equality constraints).
#'   * `EQ_DIM`: alias of `ZERO`.
#'   * `POS` (`"l"`): nonnegative orthant.
#'   * `SOC` (`"q"`): second-order cone.
#'   * `PSD` (`"s"`): symmetric positive semidefinite cone.
#'   * `EXP` (`"ep"`): primal exponential cone.
#'   * `EXP_DUAL` (`"ed"`): dual exponential cone.
#'   * `CONES`: the canonical SCS ordering of cone tags.
#' @name diffcp-cones
#' @keywords internal
NULL

#' @export
#' @rdname diffcp-cones
ZERO     <- "z"
#' @export
#' @rdname diffcp-cones
EQ_DIM   <- ZERO
#' @export
#' @rdname diffcp-cones
POS      <- "l"
#' @export
#' @rdname diffcp-cones
SOC      <- "q"
#' @export
#' @rdname diffcp-cones
PSD      <- "s"
#' @export
#' @rdname diffcp-cones
EXP      <- "ep"
#' @export
#' @rdname diffcp-cones
EXP_DUAL <- "ed"

#' @export
#' @rdname diffcp-cones
CONES <- c(ZERO, POS, SOC, PSD, EXP, EXP_DUAL)

## DIFFCP SOURCE: diffcp/cones.py @ v1.1.8 (6a78143) (Python source of truth).
## Upstream pin: see inst/UPSTREAM.dcf / upstream_info().
##
## Cone projections: each `proj_*` function takes a numeric vector `x`
## of the appropriate length and returns Pi_K(x), the projection onto
## the cone K. Dual projections Pi_{K^*} are computed via the Moreau
## decomposition Pi_{K^*}(x) = x + Pi_K(-x) for self-dual / non-self-dual
## cases as needed.

#' Parse an SCS-style cone dictionary into an ordered list
#'
#' Mirrors `diffcp.cones.parse_cone_dict`.
#'
#' @param cone_dict A named list with keys among `"z"`, `"l"`, `"q"`,
#'   `"s"`, `"ep"`, `"ed"`. Values are scalar dimensions (for `z`, `l`,
#'   `ep`, `ed`) or integer vectors of dimensions (for `q`, `s`).
#' @returns A list of `list(name, size)` pairs, in the canonical SCS
#'   cone order (`CONES`).
#' @export
parse_cone_dict <- function(cone_dict) {
  out <- list()
  for (cone in CONES) {
    if (!is.null(cone_dict[[cone]])) {
      ## C++ wrapper expects `sizes` to always be an integer vector,
      ## even for scalar-dim cones (z, l, ep, ed). Match that schema
      ## here so R and C++ paths consume the same structure.
      sz <- as.integer(cone_dict[[cone]])
      out[[length(out) + 1L]] <- list(name = cone, sizes = sz)
    }
  }
  out
}

# -- Projection onto the zero cone --------------------------------
## Pi_{0}(x) = 0 (primal); Pi_{R^n}(x) = x (dual: trivial since K^*=R^n).
.proj_zero <- function(x, dual = FALSE) {
  if (dual) x else numeric(length(x))
}

# -- Projection onto the nonneg orthant ---------------------------
## Self-dual: Pi_{R_+}(x) = max(x, 0).
.proj_nonneg <- function(x, dual = FALSE) {
  pmax(x, 0)
}

# -- Projection onto a single second-order (Lorentz) cone ---------
## Q^{n+1} = { (t, z) in R x R^n : ||z||_2 <= t }.
## Self-dual: same projection for primal and dual.
##
## DIFFCP SOURCE: diffcp/cones.py:88-97 (_proj branch for SOC).
.proj_soc_one <- function(x, dual = FALSE) {
  if (length(x) == 0L) return(x)
  t <- x[1L]
  z <- x[-1L]
  norm_z <- sqrt(sum(z * z))
  if (norm_z <= t || isTRUE(all.equal(norm_z, t, tolerance = 1e-8))) {
    return(x)
  }
  if (norm_z <= -t) {
    return(numeric(length(x)))
  }
  scale <- 0.5 * (1 + t / norm_z)
  c(scale * norm_z, scale * z)
}

# -- Projection onto a Cartesian product of SOC cones -------------
## `dims` is an integer vector; the input `x` is split into chunks of
## sizes `dims` and each is projected independently.
.proj_soc <- function(x, dims, dual = FALSE) {
  out <- numeric(length(x))
  offset <- 0L
  for (d in dims) {
    block <- x[(offset + 1L):(offset + d)]
    out[(offset + 1L):(offset + d)] <- .proj_soc_one(block, dual = dual)
    offset <- offset + d
  }
  out
}

# -- Projection onto the PSD cone ---------------------------------
## DIFFCP SOURCE: diffcp/cones.py:99-102 (PSD branch of `_proj`).
##
## PSD is self-dual, so the `dual` argument is ignored.  `x` is the
## SCS-style packed lower triangle (length n*(n+1)/2 with sqrt(2)
## scaling on off-diagonals).  We unpack to a symmetric matrix,
## clip eigenvalues to [0, +Inf), and re-pack.
.proj_psd <- function(x, dim) {
  X <- unvec_symm(x, dim)
  eig <- eigen(X, symmetric = TRUE)
  X_proj <- eig$vectors %*% (pmax(eig$values, 0) * t(eig$vectors))
  vec_symm(X_proj)
}

# -- Single-cone dispatch -----------------------------------------
## Internal helper used by `pi()` to project one block.
.proj_one_block <- function(x, cone, dual = FALSE) {
  switch(cone,
    z  = .proj_zero(x, dual),
    l  = .proj_nonneg(x, dual),
    q  = .proj_soc_one(x, dual),
    ## PSD, EXP, EXP_DUAL deferred to a subsequent commit.
    cli::cli_abort(
      "Cone {.val {cone}} not yet implemented in this development version."
    )
  )
}

#' Projection onto a Cartesian product of cones
#'
#' Mirrors `diffcp.cones.pi`.
#'
#' @param x A numeric vector.
#' @param cones A list of `list(name, size)` pairs, as produced by
#'   `parse_cone_dict`.
#' @param dual If `TRUE`, project onto the dual cones.
#' @returns A numeric vector of the same length as `x`.
#' @export
pi <- function(x, cones, dual = FALSE) {
  out <- numeric(length(x))
  offset <- 0L
  for (entry in cones) {
    cone <- entry$name
    sz   <- entry$sizes
    if (length(sz) == 0L || sum(sz) == 0L) next
    if (cone == "q") {
      ## SOC: each entry is one cone of size `d`.
      for (d in sz) {
        out[(offset + 1L):(offset + d)] <-
          .proj_soc_one(x[(offset + 1L):(offset + d)], dual = dual)
        offset <- offset + d
      }
    } else if (cone == "s") {
      ## PSD: each entry is the matrix dimension; vector size is dim*(dim+1)/2.
      for (d in sz) {
        ## NOTE %/% has higher precedence than * in R; ALWAYS parenthesize.
        sz_vec <- (d * (d + 1L)) %/% 2L
        block_idx <- (offset + 1L):(offset + sz_vec)
        out[block_idx] <- .proj_psd(x[block_idx], d)
        offset <- offset + sz_vec
      }
    } else if (cone == "ep" || cone == "ed") {
      ## DIFFCP SOURCE: diffcp/cones.py:80-114 (EXP / EXP_DUAL branch
      ## of _proj).  EXP_DUAL is handled by toggling `dual`:
      ##   EXP_DUAL, dual=F  ->  EXP, dual=T
      ##   EXP_DUAL, dual=T  ->  EXP, dual=F
      ## Within EXP itself, dual=T uses Moreau:
      ##   Pi_{K_exp^*}(x) = x + Pi_{K_exp}(-x).
      eff_dual <- if (cone == "ed") !dual else dual
      for (count in sz) {
        for (k in seq_len(count)) {
          idx <- (offset + 1L):(offset + 3L)
          xi <- x[idx]
          if (eff_dual) xi <- -xi
          proj <- as.numeric(cpp_project_exp_cone(xi))
          out[idx] <- if (eff_dual) x[idx] + proj else proj
          offset <- offset + 3L
        }
      }
    } else {
      ## Scalar-dim cones: zero, nonneg.
      for (d in sz) {
        block_idx <- (offset + 1L):(offset + d)
        out[block_idx] <- .proj_one_block(x[block_idx], cone, dual = dual)
        offset <- offset + d
      }
    }
  }
  out
}

## CVXPY parity: each test pins to numerics computed via Python
## diffcp's cones.pi / _proj on the same input vector. See
## scripts/diffcp_expected.py (in the diffcp R package source) for
## how to regenerate.

# -- Zero cone -----------------------------------------------------

test_that("Zero cone primal projection is zero", {
  expect_equal(diffcp:::.proj_zero(c(1, 2, 3)), c(0, 0, 0))
})

test_that("Zero cone dual projection is identity", {
  expect_equal(diffcp:::.proj_zero(c(1, 2, 3), dual = TRUE), c(1, 2, 3))
})

# -- Nonneg orthant ------------------------------------------------

test_that("Nonneg orthant projection clips negatives to 0", {
  expect_equal(diffcp:::.proj_nonneg(c(-1, 0, 2, -3)), c(0, 0, 2, 0))
})

# -- SOC -----------------------------------------------------------

test_that("SOC projection: interior of K is the identity", {
  ## ||z|| < t -> in the cone -> identity
  x <- c(2, 0.5, 0.5)
  expect_equal(diffcp:::.proj_soc_one(x), x)
})

test_that("SOC projection: interior of -K is zero", {
  ## ||z|| < -t -> in the polar -> zero
  x <- c(-2, 0.5, 0.5)
  expect_equal(diffcp:::.proj_soc_one(x), c(0, 0, 0))
})

test_that("SOC projection: transition region uses the alpha formula", {
  ## ||z|| > |t| -> alpha = (t + ||z||)/2; out = (alpha, alpha * z / ||z||).
  ## With t=1, z=(3, 4): ||z||=5, alpha=3, scale=alpha/||z||=0.6.
  ## out = (3, 1.8, 2.4).
  x <- c(1, 3, 4)
  expect_equal(diffcp:::.proj_soc_one(x), c(3, 1.8, 2.4), tolerance = 1e-10)
})

# -- pi() over a Cartesian product --------------------------------

test_that("pi() composes per-cone projections in SCS order", {
  ## A small mixed problem: 2 zero, 3 nonneg, 1 SOC of size 3.
  cone_dict <- list(z = 2, l = 3, q = 3)
  cones <- parse_cone_dict(cone_dict)
  x <- c(1, 2,            # zero
         -1, 0, 5,        # nonneg
         1, 3, 4)         # SOC
  expected <- c(0, 0,     # zero -> 0
                0, 0, 5,  # nonneg -> max(., 0)
                3, 1.8, 2.4)  # SOC interior of K?  No: ||(3,4)||=5 > 1
  expect_equal(pi(x, cones), expected, tolerance = 1e-10)
})

# -- vec_symm / unvec_symm round-trip -----------------------------

test_that("vec_symm and unvec_symm round-trip", {
  X <- matrix(c(1, 2, 3,
                2, 4, 5,
                3, 5, 6), 3, 3, byrow = TRUE)
  v <- vec_symm(X)
  expect_length(v, 6L)  # 3*(3+1)/2 = 6
  X2 <- unvec_symm(v, 3L)
  expect_equal(X2, X, tolerance = 1e-10)
})

## DIFFCP SOURCE: diffcp/cone_program.py:632-902 @ v1.1.8 (6a78143)
## Upstream pin: see inst/UPSTREAM.dcf / upstream_info().
##
## R-side orchestration of `solve_and_derivative`. All numerical heavy
## lifting (cone projection Jacobians, M operator, dense LDLT solve,
## matrix-free LSQR) is delegated to C++ via the wrappers in
## src/wrapper.cpp.
##
## This file mirrors the Python source line-for-line, except that the
## per-iteration linear-algebra routines (`dpi_dense`, `M_dense`,
## `_solve_derivative_dense`, `_solve_adjoint_derivative_dense`,
## `M_operator`, `lsqr`) are C++ functions reachable from R.

# -- Skewed pi onto R^n x K^* x R_+ -------------------------------
## DIFFCP SOURCE: diffcp/cone_program.py:85-92 (`pi(z, cones)`).
.skew_pi <- function(z, cones, n_x, n_y) {
  u <- z[seq_len(n_x)]
  v <- z[(n_x + 1L):(n_x + n_y)]
  w <- z[n_x + n_y + 1L]
  c(u, pi(v, cones, dual = TRUE), max(w, 0))
}

# -- Build the (N x N) Q block ------------------------------------
## DIFFCP SOURCE: diffcp/cone_program.py:676-680 (Q at z*) and
## :716-720 (dQ inside derivative()). Returns a sparse dgCMatrix.
.build_Q <- function(A, b, c) {
  m <- nrow(A); n <- ncol(A)
  N <- m + n + 1L
  ## Sparse-bmat equivalent: explicitly build (i, j, x) triplets.
  AT <- methods::as(Matrix::t(A), "TsparseMatrix")
  Acoo <- methods::as(A, "TsparseMatrix")
  ## block (1:n, (n+1):(n+m)) = A^T  (rows of A^T are A's columns)
  i_at <- AT@i + 1L
  j_at <- AT@j + 1L + n
  ## block ((n+1):(n+m), 1:n) = -A
  i_a  <- Acoo@i + 1L + n
  j_a  <- Acoo@j + 1L
  ## column N: c stacked over b stacked over 0; row N: -c^T stacked
  ## with -b^T stacked with 0 (mirroring deriv.cpp / cone_program.py).
  i_col <- c(seq_len(n), seq.int(n + 1L, n + m))
  j_col <- rep.int(N, n + m)
  x_col <- c(c, b)
  i_row <- rep.int(N, n + m)
  j_row <- c(seq_len(n), seq.int(n + 1L, n + m))
  x_row <- c(-c, -b)

  Matrix::sparseMatrix(
    i = c(i_at, i_a, i_col, i_row),
    j = c(j_at, j_a, j_col, j_row),
    x = c(AT@x, -Acoo@x, x_col, x_row),
    dims = c(N, N)
  )
}

# -- Build the closures used by `solve_and_derivative` ------------
## DIFFCP SOURCE: diffcp/cone_program.py:696-774 (derivative +
## adjoint_derivative closures).
## `pattern` is the sparsity pattern `dA` is reported on, as a
## `Matrix::summary()` data frame (1-based `i`, `j`).  The caller supplies it
## because it must be read BEFORE explicit zeros are eliminated -- see the long
## comment at the `A_pattern <- ...` line in `solve_and_derivative`
## (cone_program.R), and DIFFCP SOURCE cone_program.py:643-653.  When it is
## NULL the pattern is derived from `A` as before, so a direct caller that
## passes an already-cleaned `A` is unaffected.
.make_derivative_closures <- function(A, b, c, x, y, s, cones, mode,
                                      pattern = NULL) {
  if (!mode %in% c("dense", "lsqr")) {
    cli::cli_abort("Unsupported mode {.val {mode}}; supported: 'dense', 'lsqr'.")
  }
  m <- nrow(A); n <- ncol(A); N <- m + n + 1L

  ## The SOLVE / Q path always uses the eliminated matrix: the C++ routines
  ## (cpp_lsqr_M, cpp_M_dense) see exactly what they saw before this change.
  A_drop <- Matrix::drop0(A)
  A_csc  <- methods::as(A_drop, "CsparseMatrix")
  nz <- if (is.null(pattern)) Matrix::summary(A_csc) else pattern
  rows <- nz$i  # 1-based
  cols <- nz$j  # 1-based

  z <- c(x, y - s, 1)
  Q_sp <- .build_Q(A_csc, b, c)
  pi_z <- .skew_pi(z, cones, n_x = n, n_y = m)

  ## D pi(v) for the dual cone, materialized once via the C++ side so
  ## the same code path serves all cones (Zero/Nonneg/SOC today,
  ## PSD/EXP when those land).
  Dproj_dual <- cpp_dprojection_dense(z[(n + 1L):(n + m)], cones, dual = TRUE)

  if (mode == "dense") {
    Q_dense <- as.matrix(Q_sp)
    M  <- cpp_M_dense(Q_dense, cones, u = x, v = y - s, w = 1)
    MT <- t(M)
  }

  derivative <- function(dA, db, dc) {
    dA_csc <- methods::as(dA, "CsparseMatrix")
    dQ_sp  <- .build_Q(dA_csc, db, dc)
    rhs    <- as.numeric(dQ_sp %*% pi_z)
    if (all(abs(rhs) < .Machine$double.eps)) {
      dz <- numeric(N)
    } else if (mode == "dense") {
      dz <- as.numeric(cpp_solve_derivative_dense(M, MT, rhs))
    } else {
      Q_map <- methods::as(Q_sp, "CsparseMatrix")
      out <- cpp_lsqr_M(Q_map, cones, x, y - s, 1, rhs, transpose = FALSE)
      dz <- as.numeric(out$solution)
    }
    du <- dz[1L:n]
    dv <- dz[(n + 1L):(n + m)]
    dw <- dz[N]
    Ddv <- as.numeric(Dproj_dual %*% dv)
    dx <- du - x * dw
    dy <- Ddv - y * dw
    ds <- Ddv - dv - s * dw
    list(dx = -dx, dy = -dy, ds = -ds)
  }

  adjoint_derivative <- function(dx, dy, ds) {
    dw <- -(sum(x * dx) + sum(y * dy) + sum(s * ds))
    dz <- c(dx, as.numeric(crossprod(Dproj_dual, dy + ds)) - ds, dw)
    if (all(abs(dz) < .Machine$double.eps)) {
      r <- numeric(N)
    } else if (mode == "dense") {
      r <- as.numeric(cpp_solve_adjoint_derivative_dense(M, MT, dz))
    } else {
      Q_map <- methods::as(Q_sp, "CsparseMatrix")
      out <- cpp_lsqr_M(Q_map, cones, x, y - s, 1, dz, transpose = TRUE)
      r <- as.numeric(out$solution)
    }
    values <- pi_z[cols] * r[rows + n] - pi_z[n + rows] * r[cols]
    dA <- Matrix::sparseMatrix(i = rows, j = cols, x = values,
                               dims = c(m, n))
    db <- pi_z[(n + 1L):(n + m)] * r[N] - pi_z[N] * r[(n + 1L):(n + m)]
    dc <- pi_z[1L:n]              * r[N] - pi_z[N] * r[1L:n]
    list(dA = dA, db = db, dc = dc)
  }

  list(D = derivative, DT = adjoint_derivative)
}

## Finite-difference correctness tests, mirroring Python diffcp's
## `test_solve_and_derivative` (test_clarabel.py:9-43, test_scs.py:9-43).
##
## Idea: solve at (A, b, c), then re-solve at (A + dA, b + db, c + dc)
## for small random (dA, db, dc).  The forward derivative D should
## satisfy  D(dA, db, dc) ~= (x_pert - x, y_pert - y, s_pert - s)
## up to atol=1e-8 (the tolerance Python uses).
##
## For the adjoint: applying DT(c, 0, 0) yields directions
## (dA, db, dc) such that perturbing the problem by 1e-6 * (dA, db, dc)
## changes the objective c^T x by 1e-6 * <(dA, db, dc), (dA, db, dc)>
## (a self-consistency check on the adjoint, also at atol=1e-8).

skip_if_not_installed("clarabel")

test_that("D matches finite-difference (LP, mode='dense')", {
  set.seed(0)
  A_data <- c(1, -1, 0, 0,
              1,  0, -1, 0,
              1,  0, 0, -1)
  A <- Matrix::Matrix(matrix(A_data, nrow = 4, ncol = 3), sparse = TRUE)
  b <- c(1, 0, 0, 0)
  c <- c(1, 2, 3)
  cone_dict <- list(z = 1L, l = 3L)
  ctrl <- list(tol_gap_abs = 1e-12, tol_gap_rel = 1e-12, tol_feas = 1e-12)

  res <- do.call(diffcp::solve_and_derivative,
                 c(list(A, b, c, cone_dict, mode = "dense"), ctrl))

  ## Random perturbation with the same nonzero pattern as A.
  A_csc <- methods::as(A, "CsparseMatrix")
  nz <- Matrix::summary(A_csc)
  dA <- Matrix::sparseMatrix(i = nz$i, j = nz$j,
                             x = rnorm(nrow(nz)) * 1e-6,
                             dims = c(nrow(A), ncol(A)))
  db <- rnorm(length(b)) * 1e-6
  dc <- rnorm(length(c)) * 1e-6

  d <- res$D(dA, db, dc)

  res_pert <- do.call(diffcp::solve_only,
                      c(list(A + dA, b + db, c + dc, cone_dict), ctrl))

  ## Same atol=1e-8 as Python (test_clarabel.py:29-31).  testthat's
  ## tolerance arg is relative; use expect_lt on the max abs residual
  ## to match numpy's `np.testing.assert_allclose(..., atol=1e-8)`.
  expect_lt(max(abs(d$dx - (res_pert$x - res$x))), 1e-8)
  expect_lt(max(abs(d$dy - (res_pert$y - res$y))), 1e-8)
  expect_lt(max(abs(d$ds - (res_pert$s - res$s))), 1e-8)
})

test_that("DT objective sensitivity matches finite-difference (LP, mode='dense')", {
  A_data <- c(1, -1, 0, 0,
              1,  0, -1, 0,
              1,  0, 0, -1)
  A <- Matrix::Matrix(matrix(A_data, nrow = 4, ncol = 3), sparse = TRUE)
  b <- c(1, 0, 0, 0)
  c <- c(1, 2, 3)
  cone_dict <- list(z = 1L, l = 3L)
  ctrl <- list(tol_gap_abs = 1e-12, tol_gap_rel = 1e-12, tol_feas = 1e-12)

  res <- do.call(diffcp::solve_and_derivative,
                 c(list(A, b, c, cone_dict, mode = "dense"), ctrl))
  objective <- sum(c * res$x)

  ## DT(c, 0, 0) gives steepest descent of the objective in (A, b, c).
  da <- res$DT(c, numeric(length(b)), numeric(length(b)))

  eps <- 1e-6
  res_pert <- do.call(diffcp::solve_only,
                      c(list(A + eps * da$dA, b + eps * da$db, c + eps * da$dc,
                             cone_dict), ctrl))
  objective_pert <- sum(c * res_pert$x)

  ## Mirrors test_clarabel.py:41-43:
  ##   assert_allclose(obj_pert - obj,
  ##                   eps * (sum(dA*dA) + db@db + dc@dc),  atol=1e-8)
  dA_sumsq <- sum(da$dA * da$dA)
  expected <- eps * (dA_sumsq + sum(da$db * da$db) + sum(da$dc * da$dc))
  expect_lt(abs(objective_pert - objective - expected), 1e-8)
})

test_that("D matches finite-difference (SOC, mode='dense')", {
  set.seed(1)
  A_data <- c(0, -1, 0, 0, 0,
              1,  0, -1, 0, 0,
              1,  0, 0, -1, 0,
              1,  0, 0, 0, -1)
  A <- Matrix::Matrix(matrix(A_data, nrow = 5, ncol = 4), sparse = TRUE)
  b <- c(1, 0, 0, 0, 0)
  c <- c(1, 0, 0, 0)
  cone_dict <- list(z = 1L, q = 4L)
  ctrl <- list(tol_gap_abs = 1e-12, tol_gap_rel = 1e-12, tol_feas = 1e-12)

  res <- do.call(diffcp::solve_and_derivative,
                 c(list(A, b, c, cone_dict, mode = "dense"), ctrl))

  A_csc <- methods::as(A, "CsparseMatrix")
  nz <- Matrix::summary(A_csc)
  dA <- Matrix::sparseMatrix(i = nz$i, j = nz$j,
                             x = rnorm(nrow(nz)) * 1e-6,
                             dims = c(nrow(A), ncol(A)))
  db <- rnorm(length(b)) * 1e-6
  dc <- rnorm(length(c)) * 1e-6

  d <- res$D(dA, db, dc)
  res_pert <- do.call(diffcp::solve_only,
                      c(list(A + dA, b + db, c + dc, cone_dict), ctrl))

  expect_lt(max(abs(d$dx - (res_pert$x - res$x))), 1e-8)
  expect_lt(max(abs(d$dy - (res_pert$y - res$y))), 1e-8)
  expect_lt(max(abs(d$ds - (res_pert$s - res$s))), 1e-8)
})

test_that("D matches finite-difference (LP, mode='lsqr')", {
  set.seed(2)
  A_data <- c(1, -1, 0, 0,
              1,  0, -1, 0,
              1,  0, 0, -1)
  A <- Matrix::Matrix(matrix(A_data, nrow = 4, ncol = 3), sparse = TRUE)
  b <- c(1, 0, 0, 0)
  c <- c(1, 2, 3)
  cone_dict <- list(z = 1L, l = 3L)
  ctrl <- list(tol_gap_abs = 1e-12, tol_gap_rel = 1e-12, tol_feas = 1e-12)

  res <- do.call(diffcp::solve_and_derivative,
                 c(list(A, b, c, cone_dict, mode = "lsqr"), ctrl))

  A_csc <- methods::as(A, "CsparseMatrix")
  nz <- Matrix::summary(A_csc)
  dA <- Matrix::sparseMatrix(i = nz$i, j = nz$j,
                             x = rnorm(nrow(nz)) * 1e-6,
                             dims = c(nrow(A), ncol(A)))
  db <- rnorm(length(b)) * 1e-6
  dc <- rnorm(length(c)) * 1e-6

  d <- res$D(dA, db, dc)
  res_pert <- do.call(diffcp::solve_only,
                      c(list(A + dA, b + db, c + dc, cone_dict), ctrl))

  ## LSQR is an iterative solve; its convergence tolerance bounds the
  ## achievable agreement.  Use the same atol=1e-8 as Python (Python's
  ## test_clarabel.py uses mode="lsqr" by default and asserts 1e-8).
  expect_lt(max(abs(d$dx - (res_pert$x - res$x))), 1e-8)
  expect_lt(max(abs(d$dy - (res_pert$y - res$y))), 1e-8)
  expect_lt(max(abs(d$ds - (res_pert$s - res$s))), 1e-8)
})

## PSD-cone tests, ported from
##   /Users/naras/research/cvxr/new_design/diffing/diffcp/tests/test_clarabel_psd.py
##
## Each Python test builds a problem in cvxpy, canonicalizes to SCS form
## via `scs_data_from_cvxpy_problem`, then solves via `diffcp.solve_and_derivative`
## with both SCS and Clarabel and asserts:
##   * obj_solver ~= cvxpy_obj at atol=1e-4
##   * x_scs ~= x_clarabel at atol=1e-4
##   * (some tests) s_scs ~= s_clarabel at atol=1e-4
##   * test_constraint_satisfaction: A @ x + s ~= b at atol=1e-5
##   * test_derivative_lsqr_mode: DT() runs in lsqr mode and is finite
##
## We don't have cvxpy in R; the canonicalized (A, b, c, cone_dict)
## and `cvxpy_obj` reference come from inst/python_parity/psd.py and
## are pinned here as numeric literals.

skip_if_not_installed("clarabel")
skip_if_not_installed("scs")

## Helper: build sparse A from CSC triplets emitted by the parity script
## (1-based row indices and 1-based indptr).
.csc <- function(A_i, A_p, A_x, dims) {
  Matrix::sparseMatrix(i = A_i, p = A_p - 1L, x = A_x, dims = dims, index1 = TRUE)
}

## Helper: solve via both Clarabel and SCS, return list of solutions.
.solve_both <- function(A, b, c, cone_dict, ctrl_clarabel, ctrl_scs) {
  res_cla <- do.call(diffcp::solve_only,
                     c(list(A, b, c, cone_dict, solve_method = "Clarabel"),
                       ctrl_clarabel))
  res_scs <- do.call(diffcp::solve_only,
                     c(list(A, b, c, cone_dict, solve_method = "SCS"),
                       ctrl_scs))
  list(clarabel = res_cla, scs = res_scs)
}

# ---- Tighter tolerances than Python's atol=1e-4 ------------------
ctrl_cla <- list(tol_gap_abs = 1e-9, tol_gap_rel = 1e-9, tol_feas = 1e-9)
ctrl_scs <- list(eps_abs = 1e-9, eps_rel = 1e-9, max_iters = 100000L)

# ---- Port of test_multiple_psd_cones_objective_match -------------
test_that("multiple PSD cones: objective and x match between SCS and Clarabel", {
  A_i <- c(1, 10, 1, 11, 1, 12, 1, 13, 1, 14, 1, 15, 2, 4, 5, 6, 7, 8, 9,
           3, 4, 5, 6, 7, 8, 9)
  A_p <- c(1, 3, 5, 7, 9, 11, 13, 20, 27)
  A_x <- c(1, -1, 4, -1.414213562373095, 6, -1.414213562373095,
           4, -1, 10, -1.414213562373095, 6, -1, -1, -1,
           -2.82842712474619, -4.242640687119286, -4, -7.071067811865476, -6,
           -1, -7, -11.31370849898476, -12.72792206135786, -10,
           -15.55634918610405, -12)
  A <- .csc(A_i, A_p, A_x, c(15, 8))
  b <- c(1, rep(0, 14))
  c <- c(1, 0, 0, 1, 0, 1, 1, 1)
  cone_dict <- list(l = 2L, s = c(3L, 3L), z = 1L)
  obj_ref <- 0.08814600004013824

  sol <- .solve_both(A, b, c, cone_dict, ctrl_cla, ctrl_scs)

  obj_cla <- sum(c * sol$clarabel$x)
  obj_scs <- sum(c * sol$scs$x)
  expect_lt(abs(obj_cla - obj_ref), 1e-4)
  expect_lt(abs(obj_scs - obj_ref), 1e-4)
  expect_lt(max(abs(sol$clarabel$x - sol$scs$x)), 1e-4)
  expect_lt(max(abs(sol$clarabel$s - sol$scs$s)), 1e-4)
})

# ---- Port of test_single_psd_cone --------------------------------
test_that("single PSD cone: objective and x match between SCS and Clarabel", {
  A_i <- c(1, 2, 3, 4, 1, 5, 6, 1, 7)
  A_p <- c(1, 3, 4, 5, 7, 8, 10)
  A_x <- c(1, -1, -1.414213562373095, -1.414213562373095,
           1, -1, -1.414213562373095, 1, -1)
  A <- .csc(A_i, A_p, A_x, c(7, 6))
  b <- c(1, rep(0, 6))
  c <- c(1, 0, 0, 1, 0, 1)
  cone_dict <- list(s = c(3L), z = 1L)
  obj_ref <- 1.000000000000004

  sol <- .solve_both(A, b, c, cone_dict, ctrl_cla, ctrl_scs)
  obj_cla <- sum(c * sol$clarabel$x)
  obj_scs <- sum(c * sol$scs$x)
  expect_lt(abs(obj_cla - obj_ref), 1e-4)
  expect_lt(abs(obj_scs - obj_ref), 1e-4)
  expect_lt(max(abs(sol$clarabel$x - sol$scs$x)), 1e-4)
  expect_lt(max(abs(sol$clarabel$s - sol$scs$s)), 1e-4)
})

# ---- Port of test_mixed_cones ------------------------------------
test_that("mixed cones (z + l + PSD): objective and x match between SCS and Clarabel", {
  A_i <- c(1, 4, 1, 5, 1, 6, 2, 3)
  A_p <- c(1, 3, 5, 7, 9)
  A_x <- c(1, -1, 1, -1.414213562373095, 2, -1, -1, 1)
  A <- .csc(A_i, A_p, A_x, c(6, 4))
  b <- c(1, 0, 5, 0, 0, 0)
  c <- c(1, 0, 1, 1)
  cone_dict <- list(l = 2L, s = c(2L), z = 1L)
  obj_ref <- 0.4530818398464256

  sol <- .solve_both(A, b, c, cone_dict, ctrl_cla, ctrl_scs)
  obj_cla <- sum(c * sol$clarabel$x)
  obj_scs <- sum(c * sol$scs$x)
  expect_lt(abs(obj_cla - obj_ref), 1e-4)
  expect_lt(abs(obj_scs - obj_ref), 1e-4)
  expect_lt(max(abs(sol$clarabel$x - sol$scs$x)), 1e-4)
})

# ---- Port of test_constraint_satisfaction ------------------------
test_that("Clarabel solution satisfies A @ x + s = b for multi-PSD problem", {
  A_i <- c(1, 10, 1, 11, 1, 12, 1, 13, 1, 14, 1, 15, 2, 4, 5, 6, 7, 8, 9,
           3, 4, 5, 6, 7, 8, 9)
  A_p <- c(1, 3, 5, 7, 9, 11, 13, 20, 27)
  A_x <- c(1, -1, 4, -1.414213562373095, 6, -1.414213562373095,
           4, -1, 10, -1.414213562373095, 6, -1, -1, -1,
           -2.82842712474619, -4.242640687119286, -4, -7.071067811865476, -6,
           -1, -7, -11.31370849898476, -12.72792206135786, -10,
           -15.55634918610405, -12)
  A <- .csc(A_i, A_p, A_x, c(15, 8))
  b <- c(1, rep(0, 14))
  c <- c(1, 0, 0, 1, 0, 1, 1, 1)
  cone_dict <- list(l = 2L, s = c(3L, 3L), z = 1L)

  res <- do.call(diffcp::solve_only,
                 c(list(A, b, c, cone_dict, solve_method = "Clarabel"),
                   ctrl_cla))
  residual <- as.numeric(A %*% res$x + res$s - b)
  expect_lt(max(abs(residual)), 1e-5)
})

# ---- Port of test_derivative_lsqr_mode ---------------------------
test_that("Clarabel + lsqr mode: DT runs and returns finite values (PSD)", {
  A_i <- c(1, 2, 3, 1, 4)
  A_p <- c(1, 3, 4, 6)
  A_x <- c(1, -1, -1.414213562373095, 1, -1)
  A <- .csc(A_i, A_p, A_x, c(4, 3))
  b <- c(1, 0, 0, 0)
  c <- c(1, 0.6, 2)
  cone_dict <- list(s = c(2L), z = 1L)

  res <- do.call(diffcp::solve_and_derivative,
                 c(list(A, b, c, cone_dict,
                        solve_method = "Clarabel", mode = "lsqr"),
                   ctrl_cla))

  set.seed(42)
  dx <- rnorm(length(res$x)) * 0.01
  dy <- rnorm(length(res$y)) * 0.01
  ds <- rnorm(length(res$s)) * 0.01

  da <- res$DT(dx, dy, ds)
  expect_true(all(is.finite(da$dc)), info = "dc has non-finite values")
  expect_true(all(is.finite(da$db)), info = "db has non-finite values")
  expect_true(all(is.finite(methods::as(da$dA, "CsparseMatrix")@x)),
              info = "dA has non-finite values")
})

# ---- Port of test_psd_permutation_logic --------------------------
test_that("PSD permutation logic: random C agrees with cvxpy reference", {
  A_i <- c(1, 2, 3, 4, 1, 5, 6, 1, 7)
  A_p <- c(1, 3, 4, 5, 7, 8, 10)
  A_x <- c(1, -1, -1.414213562373095, -1.414213562373095,
           1, -1, -1.414213562373095, 1, -1)
  A <- .csc(A_i, A_p, A_x, c(7, 6))
  b <- c(1, rep(0, 6))
  c <- c(4.229934512253635, 7.487741429710518, 3.028829625895845,
         9.464447231296065, 3.894503620788191, 0.9362311369855318)
  cone_dict <- list(s = c(3L), z = 1L)
  obj_ref <- 0.306530700912452

  res <- do.call(diffcp::solve_only,
                 c(list(A, b, c, cone_dict, solve_method = "Clarabel"),
                   ctrl_cla))
  obj_cla <- sum(c * res$x)
  expect_lt(abs(obj_cla - obj_ref), 1e-4)
})

---
title: "Differentiating Through a Cone Program with diffcp"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Differentiating Through a Cone Program with diffcp}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Introduction

Suppose you have a convex optimization problem whose data — the
constraint matrix, the offset, the objective coefficients — depend
on some upstream quantity (a parameter, a model weight, a
hyperparameter, the output of a previous computation). You solve
the problem and get an optimum. **How does that optimum change
when the data changes?**

`diffcp` answers that question. It treats the optimal solution as
an *implicit function* of the problem data, and returns the
derivative of that function — both in the forward direction
(perturb the data, get the predicted change in the optimum) and as
its adjoint (start from a downstream loss on the optimum and
backpropagate to a gradient on the data).

Three concrete things you can do with it:

* **Sensitivity analysis.** Quantify how much each constraint or
  cost coefficient matters at the optimum.
* **Hyperparameter learning.** Solve a regularized problem inside
  a training loop and let gradient descent on the upstream loss
  learn the regularizer.
* **End-to-end optimization layers.** Embed a CVXR problem as a
  differentiable component of a larger model — see CVXR's
  `Problem$backward()` and `Problem$derivative()` API, which
  internally calls `diffcp`.

The implementation is a port of the Python `diffcp` package
(Agrawal, Barratt, Boyd, Busseti, Moursi, 2019); the numerical
core (cone projections, Jacobians, the `M` operator, LSQR) is
direct C++ from the upstream repository, called from R via
RcppEigen.

```{r setup}
library(diffcp)
library(Matrix)
```

# The mathematical setup

`diffcp` differentiates the *primal–dual conic problem*

$$
\begin{aligned}
\text{minimize}   &\quad c^\top x \\
\text{subject to} &\quad A x + s = b, \quad s \in K,
\end{aligned}
$$

where $K$ is a Cartesian product of the standard cones (zero, the
non-negative orthant, second-order, positive semidefinite,
exponential, and the exponential dual). The dual variable
attached to the cone constraint is $y$; together $(x, y, s)$ is
the primal–dual triple.

Differentiating "the optimum" turns out to mean something precise:

1. The KKT conditions can be lifted into a *homogeneous self-dual
   embedding* (Ye–Todd–Mizuno) in which optimality, primal
   infeasibility, and dual infeasibility all become statements
   about a single residual map $N(z) = ((Q - I) \Pi(z) + I) z$,
   where $Q$ is a skew-symmetric block of the data and $\Pi$ is
   the projection onto $\mathbb{R}^n \times K^\ast \times
   \mathbb{R}_+$.
2. At a *regular point* — informally, an optimum away from
   active-set transitions and degeneracies — the implicit-function
   theorem applies, and the optimum is a smooth function of the
   problem data.
3. The Jacobian of the optimum with respect to the data is then a
   bounded linear operator, computed by solving a linear system in
   $M = (Q - I)\,D\Pi(z) + I$ using either a dense LDLT
   factorization or a matrix-free LSQR iteration.

This is the construction in Busseti, Moursi, and Boyd (2019);
`diffcp` packages it into two callables.

# What the package computes

Calling `solve_and_derivative(A, b, c, cone_dict)` returns a list
with five entries:

* `x`, `y`, `s` — the primal–dual optimum.
* `D(dA, db, dc)` — the **forward derivative**. Given a
  perturbation of the data, returns `(dx, dy, ds)`: the predicted
  change in the optimum, to first order.
* `DT(dx, dy, ds)` — the **adjoint** of `D`. Given a perturbation
  of the optimum (or, more usefully, a gradient of a downstream
  loss with respect to the optimum), returns `(dA, db, dc)`: the
  gradient of that loss with respect to the data.

Both are functions, not matrices. The dense mode materializes the
Jacobian as an Eigen matrix and factors it once; the LSQR mode
applies the Jacobian and its transpose matrix-free, never
forming the dense object. The trade-off is per-call cost
(`dense` is faster once `M` has been factored) versus memory and
asymptotic scaling (`lsqr` is the only viable choice for very
large problems).

# Example 1: a linear program

The smallest illustrative case: pick the cheapest non-negative
mixture of three goods that sums to one,

$$
\min_{x \in \mathbb{R}^3} \; c^\top x
\quad \text{subject to} \quad
\mathbf{1}^\top x = 1,\; x \ge 0.
$$

In SCS standard form $Ax + s = b$, $s \in K = \{0\} \times
\mathbb{R}^3_+$ — one row for the equality (zero cone, $s = 0$)
followed by three rows for the non-negativity constraints (the
slack absorbs the inequality):

```{r lp-data}
A <- sparseMatrix(i = c(1, 2, 1, 3, 1, 4),
                  j = c(1, 1, 2, 2, 3, 3),
                  x = c(1, -1, 1, -1, 1, -1),
                  dims = c(4, 3))
b <- c(1, 0, 0, 0)
c <- c(1, 2, 3)
cone_dict <- list(z = 1L, l = 3L)

res <- solve_and_derivative(A, b, c, cone_dict, mode = "lsqr",
                            tol_gap_abs = 1e-10,
                            tol_gap_rel = 1e-10,
                            tol_feas    = 1e-10)
res$x
```

The optimum is $x \approx (1, 0, 0)$ — all weight on the cheapest
coordinate. Now the derivative. Suppose we slightly increase the
cost of the first coordinate, holding everything else fixed:

```{r lp-D}
dA <- sparseMatrix(i = integer(0), j = integer(0), x = numeric(0),
                   dims = c(4, 3))
db <- numeric(4)
dc <- c(0.001, 0, 0)
d <- res$D(dA, db, dc)
d$dx
```

`D` reports that `dx` is essentially zero — the optimum is robust
to a small bump in $c_1$, because the constraint $\mathbf{1}^\top
x = 1$ pins coordinate 1 against $x \ge 0$. (You'd see a
non-trivial `dy` if you printed the dual perturbation.) The
forward derivative agrees with finite-differencing the perturbed
solve to within solver tolerance:

```{r lp-fd}
res_pert <- solve_only(A, b, c + dc, cone_dict,
                       tol_gap_abs = 1e-10,
                       tol_gap_rel = 1e-10,
                       tol_feas    = 1e-10)
max(abs(d$dx - (res_pert$x - res$x)))
```

The agreement is at the level of the Clarabel solve tolerance,
not the LSQR tolerance — the per-call derivative *is* the exact
implicit-function-theorem answer at this point.

# Example 2: a second-order-cone program

A geometric example: project the simplex centroid onto a Lorentz
cone,

$$
\min_{t,\, x} \; t \quad \text{subject to} \quad \|x\|_2 \le t,\;
\mathbf{1}^\top x = 1.
$$

The variables are $(t, x_1, x_2, x_3)$. The block $(t, x)$ lives
in a single 4-dimensional second-order cone; the equality goes in
a zero cone of size 1. Since $t$ is being minimized against a
norm constraint, the optimum lies on the cone boundary
$\|x\|_2 = t$, which is exactly the kind of active constraint
that makes derivatives non-trivial.

```{r soc-data}
A <- sparseMatrix(
  i = c(2, 1, 3, 1, 4, 1, 5),
  j = c(1, 2, 2, 3, 3, 4, 4),
  x = c(-1, 1, -1, 1, -1, 1, -1),
  dims = c(5, 4))
b <- c(1, 0, 0, 0, 0)
c <- c(1, 0, 0, 0)
cone_dict <- list(z = 1L, q = 4L)

res <- solve_and_derivative(A, b, c, cone_dict, mode = "dense",
                            tol_gap_abs = 1e-10,
                            tol_gap_rel = 1e-10,
                            tol_feas    = 1e-10)
res$x
```

The optimum is $x = (1/3, 1/3, 1/3)$, $t = \sqrt{3}/3$, and the
SOC constraint $\|x\|_2 = t$ holds with equality. The cone
projection's Jacobian at a boundary point is the technical heart
of `diffcp` — it's a smoothed projector that interpolates between
the identity (interior of the cone) and zero (interior of the
polar cone). The package handles all the boundary geometry
automatically; from the R side you just call `D` and `DT`.

# Example 3: a small semidefinite program

The minimum-eigenvalue SDP: find a unit-trace PSD matrix that
minimizes $\langle C, X \rangle$ for given $C$. With
$C = \mathrm{diag}(1, 0, 0, 0, -1)$, the optimum is rank-one and
puts all weight on the eigenvector of $C$ with most-negative
eigenvalue.

PSD support requires a vectorization convention. `diffcp` uses
the SCS convention: lower-triangular column-major, with
off-diagonals scaled by $\sqrt{2}$. The package exports
`vec_symm()` and `unvec_symm()` for round-trips between symmetric
matrices and their packed vector form. The trace-coefficient row
in the equality constraint is sparse — only the diagonal entries
contribute, and we build it explicitly:

```{r sdp-data}
DIM <- 5L
n   <- DIM * (DIM + 1L) %/% 2L          # 15

trace_row <- numeric(n)
pos <- 1L
for (col in seq_len(DIM)) {
  trace_row[pos] <- 1.0
  pos <- pos + (DIM - col + 1L)         # next diagonal in column-major lower-tri
}

A <- rbind(
  Matrix(matrix(trace_row, nrow = 1L), sparse = TRUE),
  -Diagonal(n))
A <- as(A, "CsparseMatrix")
b <- c(1.0, numeric(n))

C <- matrix(0, DIM, DIM)
C[1, 1] <- 1; C[DIM, DIM] <- -1
c_vec <- vec_symm(C)

cone_dict <- list(z = 1L, s = DIM)

res <- solve_only(A, b, c_vec, cone_dict,
                  tol_gap_abs = 1e-10,
                  tol_gap_rel = 1e-10,
                  tol_feas    = 1e-10)
X <- unvec_symm(res$x, DIM)
sum(diag(X))
range(eigen(X, symmetric = TRUE, only.values = TRUE)$values)
```

`X` has trace 1 (to solver tolerance) and is PSD (smallest
eigenvalue $\ge -10^{-9}$). Behind the scenes, when the solve
goes through Clarabel — which packs PSD entries in upper-triangular
column-major order, opposite to SCS — `diffcp` permutes the rows
of $A$ and the entries of $b$ on the way in, and inverse-permutes
the dual $y$ and slack $s$ on the way out, so the user-facing
convention stays SCS throughout.

# A practical sensitivity-analysis example

Here is the canonical use case: **shadow prices**. Take a small
LP — a profit-maximizing production plan with a labor and a
materials budget — and ask, *given the optimum, by how much would
the profit change per extra unit of labor?* That number is the
dual variable on the labor constraint. `diffcp` lets you compute
the same sensitivity for *any* perturbation of the data, including
ones that have no closed-form dual interpretation.

We solve

$$
\max_{x_1, x_2 \ge 0}\; 3 x_1 + 5 x_2
\quad \text{s.t.}\quad
2 x_1 + x_2 \le 100,\;
x_1 + 3 x_2 \le 90,
$$

and ask: how does the optimum respond to a small change in the
labor-budget right-hand side, $b_1 = 100$?

We rewrite the maximization as a minimization of $-c^\top x$, and
encode the two budget inequalities and the two non-negativity
inequalities as four rows of a non-negative orthant slack
constraint:

```{r shadow-data}
A_lp <- sparseMatrix(i = c(1, 2, 3,  1, 2, 4),
                     j = c(1, 1, 1,  2, 2, 2),
                     x = c(2, 1, -1, 1, 3, -1),
                     dims = c(4, 2))
b_lp <- c(100, 90, 0, 0)
c_lp <- c(-3, -5)
cones_lp <- list(l = 4L)

res <- solve_and_derivative(A_lp, b_lp, c_lp, cones_lp, mode = "dense",
                            tol_gap_abs = 1e-10, tol_gap_rel = 1e-10,
                            tol_feas    = 1e-10)
res$x      # optimal production plan
-sum(c_lp * res$x)   # max profit
```

The optimum makes $x_1 = 42$, $x_2 = 16$, profit $= 206$. Now use
`D` to ask how the optimal $x$ would change if we got one extra
unit of labor ($\Delta b_1 = 1$):

```{r shadow-D}
dA <- sparseMatrix(i = integer(0), j = integer(0), x = numeric(0),
                   dims = c(4, 2))
db <- c(1, 0, 0, 0)
dc <- numeric(2)
d  <- res$D(dA, db, dc)
d$dx                       # change in optimal production
-sum(c_lp * d$dx)          # change in profit
```

`d$dx` says that an extra unit of labor increases $x_1$ by 0.6
and decreases $x_2$ by 0.2, and the corresponding profit change
is $0.8$ per unit of labor: the labor-budget shadow price,
recovered from `D` rather than from the dual variable $y$.

The advantage of going through `D` instead of the dual: it
generalizes. We can ask the same question about non-trivial
perturbations of $A$ (a productivity gain on a particular
input), `c` (a price change on a particular product), or any
combination. The dual variable answers only the canonical
RHS-perturbation question; `D` answers all of them through the
same single call.

`DT` answers the *opposite* question. Suppose we have a
downstream loss $L(x^\ast)$ — say, a smooth function of the
optimum with known gradient $\partial L / \partial x^\ast$. The
adjoint computes the gradient of $L$ with respect to *every*
data entry $(A, b, c)$ in one shot:

```{r shadow-DT}
## Suppose dL/dx* is given (e.g. from autodiff in a larger model).
dL_dx <- c(1.0, -0.5)
dL_dy <- numeric(length(res$y))
dL_ds <- numeric(length(res$s))
da    <- res$DT(dL_dx, dL_dy, dL_ds)
da$db                 # gradient of L w.r.t. each entry of b
da$dc                 # gradient of L w.r.t. each entry of c
```

In a training loop you'd feed `da$db` and `da$dc` straight into a
parameter update.

# Forward versus adjoint: when to use each

Both `D` and `DT` access the same underlying Jacobian, but their
cost profile differs:

* **`D(dA, db, dc)`** is right-multiplication by the Jacobian.
  Cost is one solve of the same linear system per call. Use
  `D` when you have a *small number of perturbation directions*
  (e.g., parameter sweeps) and want to forecast the effect on
  the optimum.
* **`DT(dx, dy, ds)`** is right-multiplication by the Jacobian's
  transpose. Same per-call cost, but the answer is a *gradient
  with respect to all data entries simultaneously*. Use `DT`
  when you have an upstream gradient on the optimum (typically
  from an autodiff system) and need to propagate it through the
  conic optimization.

`mode = "dense"` factors `M` once on the first call and amortizes
the cost across subsequent `D` / `DT` invocations on the same
problem, which is the right choice when you'll evaluate either
operator many times.

# Choosing `dense` versus `lsqr`

`solve_and_derivative` accepts two modes for the inner
linear-algebra step:

* **`mode = "lsqr"`** (default): matrix-free conjugate-gradient-
  style iterative solve against the `M` operator. Memory is
  $O(\text{nnz}(A))$; iteration count grows with the conditioning
  of $M$. Appropriate for large sparse problems.
* **`mode = "dense"`**: build $M$ as a dense matrix, factor with
  Eigen's LDLT, and reuse the factorization across `D` / `DT`
  calls. Cost is $O(N^3)$ once, $O(N^2)$ per call, where
  $N = m + n + 1$. Appropriate for small problems and for
  workflows that re-apply `D` / `DT` many times.

Both modes produce the same answer to within solver tolerance.
The dense path matches Python's
$(M^\top M).\mathrm{ldlt}().\mathrm{solve}(M^\top \cdot
\mathrm{rhs})$ bit-for-bit; the LSQR path agrees with the dense
path to roughly $10^{-6}$ at the default `atol = btol = 1e-8`.

# Conventions

* **PSD vectorization.** `vec_symm()` packs the lower triangle in
  column-major order with off-diagonals scaled by $\sqrt{2}$
  (the SCS convention). `unvec_symm()` is the inverse. This
  scaling makes the inner product $\langle \mathrm{vec}(A),
  \mathrm{vec}(B) \rangle$ equal $\mathrm{tr}(AB)$ for symmetric
  $A, B$, which is what the cone-projection Jacobian expects.
* **PSD with Clarabel.** Clarabel uses upper-triangular column-
  major; `diffcp` permutes rows of $A$ and entries of $b$ to
  Clarabel order on the way in and inverse-permutes the dual
  $y$ and slack $s$ on the way out. The user-facing convention
  is SCS throughout.
* **Exponential cones.** Each `ep` cone consumes three rows of
  $A$ (one cone per `count`); the dual `ed` cone is handled
  separately or recovered via Moreau's decomposition,
  $\Pi_{K^\ast}(x) = x + \Pi_K(-x)$.
* **Quadratic objectives.** `solve_only(P = P)` accepts a
  positive-semidefinite `P` for the forward solve. The
  derivative path through `solve_and_derivative` does **not**
  support `P` directly; CVXR's reduction chain canonicalizes QPs
  into auxiliary-variable conic problems before reaching the
  solver, so the loss is invisible at the CVXR layer.

# Connection to CVXR

R users typically don't call `diffcp` directly. CVXR (≥ 1.8.2)
exposes a higher-level API:

```r
library(CVXR)
p <- Parameter(); x <- Variable(); value(p) <- 3
prob <- Problem(Minimize((x - 2 * p)^2), list(x >= 0))
psolve(prob, requires_grad = TRUE)
backward(prob)
gradient(p)             # 2  (since x* = 2p)
delta(p) <- 1e-3
derivative(prob)
delta(x)                # 0.002
```

The `requires_grad = TRUE` flag routes the solve through the
DIFFCP solver wrapper, which calls `diffcp::solve_and_derivative`
under the hood. CVXR then handles the chain rule through any
problem reductions (DGP log-transform, Complex2Real splitting,
parameter canonicalization) and reports gradients keyed by
`Parameter` rather than by raw data entries.

When you want the lower-level data-perturbation interface — for
research, for fixtures that bypass CVXR canonicalization, or for
problems already in conic standard form — `diffcp` is the right
entry point.

# Citation

If you use `diffcp` in academic work, please cite **both** the R
package and the original paper. See `citation("diffcp")` for the
canonical BibTeX entries.

* **R package**

  > Narasimhan, B.; Agrawal, A.; Barratt, S.; Boyd, S.; Busseti, E.;
  > Moursi, W. *diffcp: Differentiating Through Cone Programs*.
  > R package, 2026. <https://github.com/bnaras/diffcp>

* **Original paper**

  > Agrawal, A.; Barratt, S.; Boyd, S.; Busseti, E.; Moursi, W.
  > "Differentiating through a cone program."
  > *Journal of Applied and Numerical Optimization*, **1**(2),
  > 107–115, 2019.

---
title: "Differentiating Through a Cone Program with diffcp"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Differentiating Through a Cone Program with diffcp}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# What `diffcp` is for

Suppose you have a convex optimization problem whose data — the
constraint matrix, the offset, the objective coefficients — depend
on some upstream quantity (a parameter, a model weight, a
hyperparameter, the output of a previous computation). You solve
the problem and get an optimum. **How does that optimum change
when the data changes?**

`diffcp` answers that question. It treats the optimal solution as
an *implicit function* of the problem data, and returns the
derivative of that function — both in the forward direction
(perturb the data, get the predicted change in the optimum) and as
its adjoint (start from a downstream loss on the optimum and
backpropagate to a gradient on the data).

Three concrete things you can do with it:

* **Sensitivity analysis.** Quantify how much each constraint or
  cost coefficient matters at the optimum.
* **Hyperparameter learning.** Solve a regularized problem inside
  a training loop and let gradient descent on the upstream loss
  learn the regularizer.
* **End-to-end optimization layers.** Embed a CVXR problem as a
  differentiable component of a larger model — see CVXR's
  `Problem$backward()` and `Problem$derivative()` API, which
  internally calls `diffcp`.

The implementation is a port of the Python `diffcp` package
(Agrawal, Barratt, Boyd, Busseti, Moursi, 2019); the numerical
core (cone projections, Jacobians, the `M` operator, LSQR) is
direct C++ from the upstream repository, called from R via
RcppEigen.

```{r setup}
library(diffcp)
library(Matrix)
```

# The mathematical setup

`diffcp` differentiates the *primal–dual conic problem*

$$
\begin{aligned}
\text{minimize}   &\quad c^\top x \\
\text{subject to} &\quad A x + s = b, \quad s \in K,
\end{aligned}
$$

where $K$ is a Cartesian product of the standard cones (zero, the
non-negative orthant, second-order, positive semidefinite,
exponential, and the exponential dual). The dual variable
attached to the cone constraint is $y$; together $(x, y, s)$ is
the primal–dual triple.

Differentiating "the optimum" turns out to mean something precise:

1. The KKT conditions can be lifted into a *homogeneous self-dual
   embedding* (Ye–Todd–Mizuno) in which optimality, primal
   infeasibility, and dual infeasibility all become statements
   about a single residual map $N(z) = ((Q - I) \Pi(z) + I) z$,
   where $Q$ is a skew-symmetric block of the data and $\Pi$ is
   the projection onto $\mathbb{R}^n \times K^\ast \times
   \mathbb{R}_+$.
2. At a *regular point* — informally, an optimum away from
   active-set transitions and degeneracies — the implicit-function
   theorem applies, and the optimum is a smooth function of the
   problem data.
3. The Jacobian of the optimum with respect to the data is then a
   bounded linear operator, computed by solving a linear system in
   $M = (Q - I)\,D\Pi(z) + I$ using either a dense LDLT
   factorization or a matrix-free LSQR iteration.

This is the construction in Busseti, Moursi, and Boyd (2019);
`diffcp` packages it into two callables.

# What the package computes

Calling `solve_and_derivative(A, b, c, cone_dict)` returns a list
with five entries:

* `x`, `y`, `s` — the primal–dual optimum.
* `D(dA, db, dc)` — the **forward derivative**. Given a
  perturbation of the data, returns `(dx, dy, ds)`: the predicted
  change in the optimum, to first order.
* `DT(dx, dy, ds)` — the **adjoint** of `D`. Given a perturbation
  of the optimum (or, more usefully, a gradient of a downstream
  loss with respect to the optimum), returns `(dA, db, dc)`: the
  gradient of that loss with respect to the data.

Both are functions, not matrices. The dense mode materializes the
Jacobian as an Eigen matrix and factors it once; the LSQR mode
applies the Jacobian and its transpose matrix-free, never
forming the dense object. The trade-off is per-call cost
(`dense` is faster once `M` has been factored) versus memory and
asymptotic scaling (`lsqr` is the only viable choice for very
large problems).

# Worked example 1: a linear program

The smallest illustrative case: pick the cheapest non-negative
mixture of three goods that sums to one,

$$
\min_{x \in \mathbb{R}^3} \; c^\top x
\quad \text{subject to} \quad
\mathbf{1}^\top x = 1,\; x \ge 0.
$$

In SCS standard form $Ax + s = b$, $s \in K = \{0\} \times
\mathbb{R}^3_+$ — one row for the equality (zero cone, $s = 0$)
followed by three rows for the non-negativity constraints (the
slack absorbs the inequality):

```{r lp-data}
A <- sparseMatrix(i = c(1, 2, 1, 3, 1, 4),
                  j = c(1, 1, 2, 2, 3, 3),
                  x = c(1, -1, 1, -1, 1, -1),
                  dims = c(4, 3))
b <- c(1, 0, 0, 0)
c <- c(1, 2, 3)
cone_dict <- list(z = 1L, l = 3L)

res <- solve_and_derivative(A, b, c, cone_dict, mode = "lsqr",
                            tol_gap_abs = 1e-10,
                            tol_gap_rel = 1e-10,
                            tol_feas    = 1e-10)
res$x
```

The optimum is $x \approx (1, 0, 0)$ — all weight on the cheapest
coordinate. Now the derivative. Suppose we slightly increase the
cost of the first coordinate, holding everything else fixed:

```{r lp-D}
dA <- sparseMatrix(i = integer(0), j = integer(0), x = numeric(0),
                   dims = c(4, 3))
db <- numeric(4)
dc <- c(0.001, 0, 0)
d <- res$D(dA, db, dc)
d$dx
```

`D` reports that `dx` is essentially zero — the optimum is robust
to a small bump in $c_1$, because the constraint $\mathbf{1}^\top
x = 1$ pins coordinate 1 against $x \ge 0$. (You'd see a
non-trivial `dy` if you printed the dual perturbation.) The
forward derivative agrees with finite-differencing the perturbed
solve to within solver tolerance:

```{r lp-fd}
res_pert <- solve_only(A, b, c + dc, cone_dict,
                       tol_gap_abs = 1e-10,
                       tol_gap_rel = 1e-10,
                       tol_feas    = 1e-10)
max(abs(d$dx - (res_pert$x - res$x)))
```

The agreement is at the level of the Clarabel solve tolerance,
not the LSQR tolerance — the per-call derivative *is* the exact
implicit-function-theorem answer at this point.

# Worked example 2: a second-order-cone program

A geometric example: project the simplex centroid onto a Lorentz
cone,

$$
\min_{t,\, x} \; t \quad \text{subject to} \quad \|x\|_2 \le t,\;
\mathbf{1}^\top x = 1.
$$

The variables are $(t, x_1, x_2, x_3)$. The block $(t, x)$ lives
in a single 4-dimensional second-order cone; the equality goes in
a zero cone of size 1. Since $t$ is being minimized against a
norm constraint, the optimum lies on the cone boundary
$\|x\|_2 = t$, which is exactly the kind of active constraint
that makes derivatives non-trivial.

```{r soc-data}
A <- sparseMatrix(
  i = c(2, 1, 3, 1, 4, 1, 5),
  j = c(1, 2, 2, 3, 3, 4, 4),
  x = c(-1, 1, -1, 1, -1, 1, -1),
  dims = c(5, 4))
b <- c(1, 0, 0, 0, 0)
c <- c(1, 0, 0, 0)
cone_dict <- list(z = 1L, q = 4L)

res <- solve_and_derivative(A, b, c, cone_dict, mode = "dense",
                            tol_gap_abs = 1e-10,
                            tol_gap_rel = 1e-10,
                            tol_feas    = 1e-10)
res$x
```

The optimum is $x = (1/3, 1/3, 1/3)$, $t = \sqrt{3}/3$, and the
SOC constraint $\|x\|_2 = t$ holds with equality. The cone
projection's Jacobian at a boundary point is the technical heart
of `diffcp` — it's a smoothed projector that interpolates between
the identity (interior of the cone) and zero (interior of the
polar cone). The package handles all the boundary geometry
automatically; from the R side you just call `D` and `DT`.

# Worked example 3: a small semidefinite program

The minimum-eigenvalue SDP: find a unit-trace PSD matrix that
minimizes $\langle C, X \rangle$ for given $C$. With
$C = \mathrm{diag}(1, 0, 0, 0, -1)$, the optimum is rank-one and
puts all weight on the eigenvector of $C$ with most-negative
eigenvalue.

PSD support requires a vectorization convention. `diffcp` uses
the SCS convention: lower-triangular column-major, with
off-diagonals scaled by $\sqrt{2}$. The package exports
`vec_symm()` and `unvec_symm()` for round-trips between symmetric
matrices and their packed vector form. The trace-coefficient row
in the equality constraint is sparse — only the diagonal entries
contribute, and we build it explicitly:

```{r sdp-data}
DIM <- 5L
n   <- DIM * (DIM + 1L) %/% 2L          # 15

trace_row <- numeric(n)
pos <- 1L
for (col in seq_len(DIM)) {
  trace_row[pos] <- 1.0
  pos <- pos + (DIM - col + 1L)         # next diagonal in column-major lower-tri
}

A <- rbind(
  Matrix(matrix(trace_row, nrow = 1L), sparse = TRUE),
  -Diagonal(n))
A <- as(A, "CsparseMatrix")
b <- c(1.0, numeric(n))

C <- matrix(0, DIM, DIM)
C[1, 1] <- 1; C[DIM, DIM] <- -1
c_vec <- vec_symm(C)

cone_dict <- list(z = 1L, s = DIM)

res <- solve_only(A, b, c_vec, cone_dict,
                  tol_gap_abs = 1e-10,
                  tol_gap_rel = 1e-10,
                  tol_feas    = 1e-10)
X <- unvec_symm(res$x, DIM)
sum(diag(X))
range(eigen(X, symmetric = TRUE, only.values = TRUE)$values)
```

`X` has trace 1 (to solver tolerance) and is PSD (smallest
eigenvalue $\ge -10^{-9}$). Behind the scenes, when the solve
goes through Clarabel — which packs PSD entries in upper-triangular
column-major order, opposite to SCS — `diffcp` permutes the rows
of $A$ and the entries of $b$ on the way in, and inverse-permutes
the dual $y$ and slack $s$ on the way out, so the user-facing
convention stays SCS throughout.

# A practical sensitivity-analysis example

Here is the canonical use case: **shadow prices**. Take a small
LP — a profit-maximizing production plan with a labor and a
materials budget — and ask, *given the optimum, by how much would
the profit change per extra unit of labor?* That number is the
dual variable on the labor constraint. `diffcp` lets you compute
the same sensitivity for *any* perturbation of the data, including
ones that have no closed-form dual interpretation.

We solve

$$
\max_{x_1, x_2 \ge 0}\; 3 x_1 + 5 x_2
\quad \text{s.t.}\quad
2 x_1 + x_2 \le 100,\;
x_1 + 3 x_2 \le 90,
$$

and ask: how does the optimum respond to a small change in the
labor-budget right-hand side, $b_1 = 100$?

We rewrite the maximization as a minimization of $-c^\top x$, and
encode the two budget inequalities and the two non-negativity
inequalities as four rows of a non-negative orthant slack
constraint:

```{r shadow-data}
A_lp <- sparseMatrix(i = c(1, 2, 3,  1, 2, 4),
                     j = c(1, 1, 1,  2, 2, 2),
                     x = c(2, 1, -1, 1, 3, -1),
                     dims = c(4, 2))
b_lp <- c(100, 90, 0, 0)
c_lp <- c(-3, -5)
cones_lp <- list(l = 4L)

res <- solve_and_derivative(A_lp, b_lp, c_lp, cones_lp, mode = "dense",
                            tol_gap_abs = 1e-10, tol_gap_rel = 1e-10,
                            tol_feas    = 1e-10)
res$x      # optimal production plan
-sum(c_lp * res$x)   # max profit
```

The optimum makes $x_1 = 42$, $x_2 = 16$, profit $= 206$. Now use
`D` to ask how the optimal $x$ would change if we got one extra
unit of labor ($\Delta b_1 = 1$):

```{r shadow-D}
dA <- sparseMatrix(i = integer(0), j = integer(0), x = numeric(0),
                   dims = c(4, 2))
db <- c(1, 0, 0, 0)
dc <- numeric(2)
d  <- res$D(dA, db, dc)
d$dx                       # change in optimal production
-sum(c_lp * d$dx)          # change in profit
```

`d$dx` says that an extra unit of labor increases $x_1$ by 0.6
and decreases $x_2$ by 0.2, and the corresponding profit change
is $0.8$ per unit of labor: the labor-budget shadow price,
recovered from `D` rather than from the dual variable $y$.

The advantage of going through `D` instead of the dual: it
generalizes. We can ask the same question about non-trivial
perturbations of $A$ (a productivity gain on a particular
input), `c` (a price change on a particular product), or any
combination. The dual variable answers only the canonical
RHS-perturbation question; `D` answers all of them through the
same single call.

`DT` answers the *opposite* question. Suppose we have a
downstream loss $L(x^\ast)$ — say, a smooth function of the
optimum with known gradient $\partial L / \partial x^\ast$. The
adjoint computes the gradient of $L$ with respect to *every*
data entry $(A, b, c)$ in one shot:

```{r shadow-DT}
## Suppose dL/dx* is given (e.g. from autodiff in a larger model).
dL_dx <- c(1.0, -0.5)
dL_dy <- numeric(length(res$y))
dL_ds <- numeric(length(res$s))
da    <- res$DT(dL_dx, dL_dy, dL_ds)
da$db                 # gradient of L w.r.t. each entry of b
da$dc                 # gradient of L w.r.t. each entry of c
```

In a training loop you'd feed `da$db` and `da$dc` straight into a
parameter update.

# Forward versus adjoint: when to use each

Both `D` and `DT` access the same underlying Jacobian, but their
cost profile differs:

* **`D(dA, db, dc)`** is right-multiplication by the Jacobian.
  Cost is one solve of the same linear system per call. Use
  `D` when you have a *small number of perturbation directions*
  (e.g., parameter sweeps) and want to forecast the effect on
  the optimum.
* **`DT(dx, dy, ds)`** is right-multiplication by the Jacobian's
  transpose. Same per-call cost, but the answer is a *gradient
  with respect to all data entries simultaneously*. Use `DT`
  when you have an upstream gradient on the optimum (typically
  from an autodiff system) and need to propagate it through the
  conic optimization.

`mode = "dense"` factors `M` once on the first call and amortizes
the cost across subsequent `D` / `DT` invocations on the same
problem, which is the right choice when you'll evaluate either
operator many times.

# Choosing `dense` versus `lsqr`

`solve_and_derivative` accepts two modes for the inner
linear-algebra step:

* **`mode = "lsqr"`** (default): matrix-free conjugate-gradient-
  style iterative solve against the `M` operator. Memory is
  $O(\text{nnz}(A))$; iteration count grows with the conditioning
  of $M$. Appropriate for large sparse problems.
* **`mode = "dense"`**: build $M$ as a dense matrix, factor with
  Eigen's LDLT, and reuse the factorization across `D` / `DT`
  calls. Cost is $O(N^3)$ once, $O(N^2)$ per call, where
  $N = m + n + 1$. Appropriate for small problems and for
  workflows that re-apply `D` / `DT` many times.

Both modes produce the same answer to within solver tolerance.
The dense path matches Python's
$(M^\top M).\mathrm{ldlt}().\mathrm{solve}(M^\top \cdot
\mathrm{rhs})$ bit-for-bit; the LSQR path agrees with the dense
path to roughly $10^{-6}$ at the default `atol = btol = 1e-8`.

# Conventions

* **PSD vectorization.** `vec_symm()` packs the lower triangle in
  column-major order with off-diagonals scaled by $\sqrt{2}$
  (the SCS convention). `unvec_symm()` is the inverse. This
  scaling makes the inner product $\langle \mathrm{vec}(A),
  \mathrm{vec}(B) \rangle$ equal $\mathrm{tr}(AB)$ for symmetric
  $A, B$, which is what the cone-projection Jacobian expects.
* **PSD with Clarabel.** Clarabel uses upper-triangular column-
  major; `diffcp` permutes rows of $A$ and entries of $b$ to
  Clarabel order on the way in and inverse-permutes the dual
  $y$ and slack $s$ on the way out. The user-facing convention
  is SCS throughout.
* **Exponential cones.** Each `ep` cone consumes three rows of
  $A$ (one cone per `count`); the dual `ed` cone is handled
  separately or recovered via Moreau's decomposition,
  $\Pi_{K^\ast}(x) = x + \Pi_K(-x)$.
* **Quadratic objectives.** `solve_only(P = P)` accepts a
  positive-semidefinite `P` for the forward solve. The
  derivative path through `solve_and_derivative` does **not**
  support `P` directly; CVXR's reduction chain canonicalizes QPs
  into auxiliary-variable conic problems before reaching the
  solver, so the loss is invisible at the CVXR layer.

# Connection to CVXR

R users typically don't call `diffcp` directly. CVXR (≥ 1.8.2)
exposes a higher-level API:

```r
library(CVXR)
p <- Parameter(); x <- Variable(); value(p) <- 3
prob <- Problem(Minimize((x - 2 * p)^2), list(x >= 0))
psolve(prob, requires_grad = TRUE)
backward(prob)
gradient(p)             # 2  (since x* = 2p)
delta(p) <- 1e-3
derivative(prob)
delta(x)                # 0.002
```

The `requires_grad = TRUE` flag routes the solve through the
DIFFCP solver wrapper, which calls `diffcp::solve_and_derivative`
under the hood. CVXR then handles the chain rule through any
problem reductions (DGP log-transform, Complex2Real splitting,
parameter canonicalization) and reports gradients keyed by
`Parameter` rather than by raw data entries.

When you want the lower-level data-perturbation interface — for
research, for fixtures that bypass CVXR canonicalization, or for
problems already in conic standard form — `diffcp` is the right
entry point.

# Citation

If you use `diffcp` in academic work, please cite **both** the R
package and the original paper. See `citation("diffcp")` for the
canonical BibTeX entries.

* **R package**

  > Narasimhan, B.; Agrawal, A.; Barratt, S.; Boyd, S.; Busseti, E.;
  > Moursi, W. *diffcp: Differentiating Through Cone Programs*.
  > R package, 2026. <https://github.com/bnaras/diffcp>

* **Original paper**

  > Agrawal, A.; Barratt, S.; Boyd, S.; Busseti, E.; Moursi, W.
  > "Differentiating through a cone program."
  > *Journal of Applied and Numerical Optimization*, **1**(2),
  > 107–115, 2019.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cones.R
\name{pi}
\alias{pi}
\title{Projection onto a Cartesian product of cones}
\usage{
pi(x, cones, dual = FALSE)
}
\arguments{
\item{x}{A numeric vector.}

\item{cones}{A list of \code{list(name, size)} pairs, as produced by
\code{parse_cone_dict}.}

\item{dual}{If \code{TRUE}, project onto the dual cones.}
}
\value{
A numeric vector of the same length as \code{x}.
}
\description{
Mirrors \code{diffcp.cones.pi}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diffcp-package.R
\docType{data}
\name{diffcp-cones}
\alias{diffcp-cones}
\alias{ZERO}
\alias{EQ_DIM}
\alias{POS}
\alias{SOC}
\alias{PSD}
\alias{EXP}
\alias{EXP_DUAL}
\alias{CONES}
\title{Cone tag constants}
\format{
Character scalars (and one character vector, \code{CONES}).

An object of class \code{character} of length 1.

An object of class \code{character} of length 1.

An object of class \code{character} of length 1.

An object of class \code{character} of length 1.

An object of class \code{character} of length 1.

An object of class \code{character} of length 1.

An object of class \code{character} of length 1.

An object of class \code{character} of length 6.
}
\usage{
ZERO

EQ_DIM

POS

SOC

PSD

EXP

EXP_DUAL

CONES
}
\description{
String keys matching the SCS / Clarabel cone conventions. Used as
names in the \code{cone_dict} argument to \code{\link[=solve_only]{solve_only()}} and
\code{\link[=solve_and_derivative]{solve_and_derivative()}}, and in the \verb{(name, size)} pairs returned
by \code{\link[=parse_cone_dict]{parse_cone_dict()}}.
}
\details{
\itemize{
\item \code{ZERO} (\code{"z"}): zero cone (equality constraints).
\item \code{EQ_DIM}: alias of \code{ZERO}.
\item \code{POS} (\code{"l"}): nonnegative orthant.
\item \code{SOC} (\code{"q"}): second-order cone.
\item \code{PSD} (\code{"s"}): symmetric positive semidefinite cone.
\item \code{EXP} (\code{"ep"}): primal exponential cone.
\item \code{EXP_DUAL} (\code{"ed"}): dual exponential cone.
\item \code{CONES}: the canonical SCS ordering of cone tags.
}
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cones.R
\name{parse_cone_dict}
\alias{parse_cone_dict}
\title{Parse an SCS-style cone dictionary into an ordered list}
\usage{
parse_cone_dict(cone_dict)
}
\arguments{
\item{cone_dict}{A named list with keys among \code{"z"}, \code{"l"}, \code{"q"},
\code{"s"}, \code{"ep"}, \code{"ed"}. Values are scalar dimensions (for \code{z}, \code{l},
\code{ep}, \code{ed}) or integer vectors of dimensions (for \code{q}, \code{s}).}
}
\value{
A list of \code{list(name, size)} pairs, in the canonical SCS
cone order (\code{CONES}).
}
\description{
Mirrors \code{diffcp.cones.parse_cone_dict}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{unvec_symm}
\alias{unvec_symm}
\title{Inverse of \code{vec_symm}}
\usage{
unvec_symm(x, dim)
}
\arguments{
\item{x}{A numeric vector of length \code{n*(n+1)/2}.}

\item{dim}{The matrix dimension \code{n}.}
}
\value{
The corresponding \verb{n x n} symmetric matrix.
}
\description{
Inverse of \code{vec_symm}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{vec_symm}
\alias{vec_symm}
\title{Symmetric vectorization (SCS convention)}
\usage{
vec_symm(X)
}
\arguments{
\item{X}{A symmetric matrix.}
}
\value{
A numeric vector of length \code{n*(n+1)/2}.
}
\description{
Returns a vectorized representation of a symmetric matrix \code{X},
with off-diagonal entries scaled by sqrt(2) to make the SCS-style
dot product \verb{<vec_symm(A), vec_symm(B)> = trace(A B)} hold.
}
\details{
Mirrors \code{diffcp.cones.vec_symm} (Python).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/upstream.R
\name{upstream_info}
\alias{upstream_info}
\title{Upstream Python diffcp pin}
\usage{
upstream_info()
}
\value{
A named character vector with one element per DCF field
(\code{Upstream}, \code{Version}, \code{Commit}, \code{ShortCommit}, \code{Date}, \code{URL},
\code{CVXPY}, \code{Snapshot}, \code{Notes}).
}
\description{
Returns the upstream pin metadata for this R port: which
\code{cvxgrp/diffcp} version / commit the R sources track, the date of
that commit, and diffcp's own CVXPY runtime constraint. The
authoritative record lives in \code{inst/UPSTREAM.dcf} and is read at
call time.
}
\examples{
upstream_info()
upstream_info()[["Version"]]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diffcp-package.R
\docType{package}
\name{diffcp-package}
\alias{diffcp-package}
\alias{diffcp}
\title{diffcp: Differentiating Through Cone Programs}
\description{
R port of the Python \code{diffcp} package, mirroring Agrawal, Barratt,
Boyd, Busseti, and Moursi (2019).  Computes the derivative of the
optimal solution map of a convex cone program with respect to its
data, via the implicit-function theorem on the homogeneous self-dual
embedding.
}
\details{
If you use this package in published work, please cite \emph{both} the
R package and the original paper; see \code{citation("diffcp")} for the
full pair of BibTeX entries.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/bnaras/diffcp}
  \item Report bugs at \url{https://github.com/bnaras/diffcp/issues}
}

}
\author{
\strong{Maintainer}: Balasubramanian Narasimhan \email{naras@stanford.edu}

Authors:
\itemize{
  \item Akshay Agrawal
  \item Shane Barratt
  \item Stephen Boyd
  \item Enzo Busseti
  \item Walaa Moursi
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cone_program.R
\name{solve_only}
\alias{solve_only}
\title{Solve a cone program (forward only)}
\usage{
solve_only(
  A,
  b,
  c,
  cone_dict,
  warm_start = NULL,
  solve_method = "Clarabel",
  P = NULL,
  ...
)
}
\arguments{
\item{A}{A sparse \code{dgCMatrix} constraint matrix.}

\item{b}{A numeric offset vector.}

\item{c}{A numeric objective coefficient vector.}

\item{cone_dict}{A named list with cone sizes (keys among
\code{"z"}, \code{"l"}, \code{"q"}, \code{"s"}, \code{"ep"}, \code{"ed"}).}

\item{warm_start}{Optional warm-start \code{list(x, y, s)}.}

\item{solve_method}{One of \code{"Clarabel"} (default) or \code{"SCS"}.}

\item{P}{Optional sparse \code{dgCMatrix} for QP objective.}

\item{...}{Additional control parameters forwarded to the solver.}
}
\value{
A list with elements \code{x}, \code{y}, \code{s}, \code{info}.
}
\description{
Mirrors \code{diffcp.cone_program.solve_only}. Solves
minimize  c^T x  (+ 0.5 x^T P x)
subject to  A x + s = b,   s in K
and returns the optimal \verb{(x, y, s)}. Unlike \code{solve_and_derivative}
this function does not build the derivative closures.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cone_program.R
\name{solve_and_derivative}
\alias{solve_and_derivative}
\title{Solve a cone program and return forward / adjoint derivative operators}
\usage{
solve_and_derivative(
  A,
  b,
  c,
  cone_dict,
  P = NULL,
  solve_method = "Clarabel",
  mode = "lsqr",
  warm_start = NULL,
  ...
)
}
\arguments{
\item{A}{A sparse \code{dgCMatrix} constraint matrix.}

\item{b}{A numeric offset vector.}

\item{c}{A numeric objective coefficient vector.}

\item{cone_dict}{A named list with cone sizes (keys among
\code{"z"}, \code{"l"}, \code{"q"}, \code{"s"}, \code{"ep"}, \code{"ed"}).}

\item{P}{Optional sparse \code{dgCMatrix} for QP objective.}

\item{solve_method}{One of \code{"Clarabel"} (default) or \code{"SCS"}.}

\item{mode}{Differentiation mode: \code{"lsqr"} (default), \code{"dense"},
or \code{"lpgd"}.}

\item{warm_start}{Optional warm-start \code{list(x, y, s)}.}

\item{...}{Additional control parameters forwarded to the solver.}
}
\value{
A list with elements \code{x}, \code{y}, \code{s}, \code{info}, \code{D}, \code{DT}.
\code{info} is the solver-status block returned by the underlying
solver (\code{status}, \code{iter}, \code{solveTime}, \code{pobj}, ...); see
\code{solve_only} for the same shape.
}
\description{
Mirrors \code{diffcp.cone_program.solve_and_derivative}. Solves
minimize  c^T x        s.t.  A x + s = b,  s in K
(with optional QP \verb{0.5 x^T P x} term) and returns the optimal
\verb{(x, y, s)} together with closures \code{D} (forward) and \code{DT} (adjoint)
that map perturbations of \verb{(A, b, c, [P])} to perturbations of
\verb{(x, y, s)} and vice versa.
}
